from .augmentations.add_background_noise import AddBackgroundNoise
from .augmentations.add_gaussian_noise import AddGaussianNoise
from .augmentations.add_gaussian_snr import AddGaussianSNR
from .augmentations.add_color_noise import AddColorNoise, NOISE_COLOR_DECAYS
from .augmentations.add_short_noises import AddShortNoises
from .augmentations.adjust_duration import AdjustDuration
from .augmentations.air_absorption import AirAbsorption
from .augmentations.aliasing import Aliasing
from .augmentations.apply_impulse_response import ApplyImpulseResponse
from .augmentations.band_pass_filter import BandPassFilter
from .augmentations.band_stop_filter import BandStopFilter
from .augmentations.bit_crush import BitCrush
from .augmentations.clip import Clip
from .augmentations.clipping_distortion import ClippingDistortion
from .augmentations.gain import Gain
from .augmentations.gain_transition import GainTransition
from .augmentations.high_pass_filter import HighPassFilter
from .augmentations.high_shelf_filter import HighShelfFilter
from .augmentations.lambda_transform import Lambda
from .augmentations.limiter import Limiter
from .augmentations.loudness_normalization import LoudnessNormalization
from .augmentations.low_pass_filter import LowPassFilter
from .augmentations.low_shelf_filter import LowShelfFilter
from .augmentations.mp3_compression import Mp3Compression
from .augmentations.normalize import Normalize
from .augmentations.padding import Padding
from .augmentations.peaking_filter import PeakingFilter
from .augmentations.pitch_shift import PitchShift
from .augmentations.polarity_inversion import PolarityInversion
from .augmentations.repeat_part import RepeatPart
from .augmentations.resample import Resample
from .augmentations.reverse import Reverse
from .augmentations.room_simulator import RoomSimulator
from .augmentations.seven_band_parametric_eq import SevenBandParametricEQ
from .augmentations.shift import Shift
from .augmentations.tanh_distortion import TanhDistortion
from .augmentations.time_mask import TimeMask
from .augmentations.time_stretch import TimeStretch
from .augmentations.trim import Trim
from .core.composition import Compose, OneOf, SomeOf

__version__ = "0.41.0"



from audiomentations import Compose, GainTransition, Shift


def test_print_compose():
    augmenter = Compose([GainTransition(), Shift()])
    assert (
        str(augmenter)
        == """Compose([
  GainTransition(p=0.5, min_gain_db=-24.0, max_gain_db=6.0, min_duration=0.2, max_duration=6.0, duration_unit='seconds'),
  Shift(p=0.5, min_shift=-0.5, max_shift=0.5, shift_unit='fraction', rollover=True, fade_duration=0.005),
], p=1.0)"""
    )



import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class PolarityInversion(BaseWaveformTransform):
    """
    Flip the audio samples upside-down, reversing their polarity. In other words, multiply the
    waveform by -1, so negative values become positive, and vice versa. The result will sound
    the same compared to the original when played back in isolation. However, when mixed with
    other audio sources, the result may be different. This waveform inversion technique
    is sometimes used for audio cancellation or obtaining the difference between two waveforms.
    In the context of audio data augmentation, this transform can be useful when
    training phase-aware machine learning models.
    """

    supports_multichannel = True

    def __init__(self, p: float = 0.5):
        """
        :param p: The probability of applying this transform
        """
        super().__init__(p)

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        return -samples



import numpy as np
import pytest

from audiomentations import TanhDistortion
from audiomentations.core.utils import calculate_rms


def test_single_channel():
    samples = np.random.normal(0, 0.1, size=(20480,)).astype(np.float32)
    sample_rate = 16000
    augmenter = TanhDistortion(min_distortion=0.2, max_distortion=0.6, p=1.0)

    distorted_samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == distorted_samples.dtype
    assert samples.shape == distorted_samples.shape
    assert np.amax(distorted_samples) < np.amax(samples)
    assert calculate_rms(distorted_samples) == pytest.approx(
        calculate_rms(samples), abs=1e-3
    )


def test_multichannel():
    num_channels = 3
    samples = np.random.normal(0, 0.1, size=(num_channels, 5555)).astype(np.float32)
    sample_rate = 16000
    augmenter = TanhDistortion(min_distortion=0.05, max_distortion=0.6, p=1.0)

    distorted_samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == distorted_samples.dtype
    assert samples.shape == distorted_samples.shape
    for i in range(num_channels):
        assert not np.allclose(samples[i], distorted_samples[i])
        assert calculate_rms(distorted_samples[i]) == pytest.approx(
            calculate_rms(samples[i]), abs=1e-3
        )



import numpy as np
from numpy.testing import assert_array_almost_equal

from audiomentations import Reverse


def test_single_channel():
    samples = np.array([0.5, 0.6, -0.2, 0.0], dtype=np.float32)
    sample_rate = 16000
    augmenter = Reverse(p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == np.float32
    assert samples.shape[-1] == 4


def test_multichannel():
    samples = np.array(
        [[0.9, 0.5, -0.25, -0.125, 0.0], [0.95, 0.5, -0.25, -0.125, 0.0]],
        dtype=np.float32,
    )
    sample_rate = 16000
    augmenter = Reverse(p=1.0)
    reversed_samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert reversed_samples.dtype == np.float32
    assert_array_almost_equal(
        reversed_samples,
        np.array(
            [[0.0, -0.125, -0.25, 0.5, 0.9], [0.0, -0.125, -0.25, 0.5, 0.95]],
            dtype=np.float32,
        ),
    )



from audiomentations import Normalize


def test_freeze_and_unfreeze_parameters():
    normalizer = Normalize(p=1.0)

    assert normalizer.are_parameters_frozen == False

    normalizer.freeze_parameters()
    assert normalizer.are_parameters_frozen == True

    normalizer.unfreeze_parameters()
    assert normalizer.are_parameters_frozen == False



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from audiomentations import Shift


def test_shift_fraction():
    samples = np.array([1.0, 0.5, 0.25, 0.125], dtype=np.float32)
    sample_rate = 16000

    forward_augmenter = Shift(min_shift=0.5, max_shift=0.5, fade_duration=0.0, p=1.0)
    forward_shifted_samples = forward_augmenter(
        samples=samples, sample_rate=sample_rate
    )
    assert_almost_equal(
        forward_shifted_samples, np.array([0.25, 0.125, 1.0, 0.5], dtype=np.float32)
    )
    assert forward_shifted_samples.dtype == np.float32
    assert len(forward_shifted_samples) == 4

    backward_augmenter = Shift(
        min_shift=-0.25, max_shift=-0.25, fade_duration=0.0, p=1.0
    )
    backward_shifted_samples = backward_augmenter(
        samples=samples, sample_rate=sample_rate
    )
    assert_almost_equal(
        backward_shifted_samples,
        np.array([0.5, 0.25, 0.125, 1.0], dtype=np.float32),
    )
    assert backward_shifted_samples.dtype == np.float32
    assert len(forward_shifted_samples) == 4


def test_shift_samples():
    samples = np.array([1.0, 0.5, 0.25, 0.125], dtype=np.float32)
    sample_rate = 16000

    forward_augmenter = Shift(
        min_shift=1, max_shift=1, shift_unit="samples", fade_duration=0.0, p=1.0
    )
    forward_shifted_samples = forward_augmenter(
        samples=samples, sample_rate=sample_rate
    )
    assert_almost_equal(
        forward_shifted_samples, np.array([0.125, 1.0, 0.5, 0.25], dtype=np.float32)
    )
    assert forward_shifted_samples.dtype == np.float32
    assert len(forward_shifted_samples) == 4


def test_shift_seconds():
    samples = np.array([1.0, 0.5, 0.25, 0.125], dtype=np.float32)
    sample_rate = 2

    forward_augmenter = Shift(
        min_shift=1.0, max_shift=1.0, shift_unit="seconds", fade_duration=0.0, p=1.0
    )
    forward_shifted_samples = forward_augmenter(
        samples=samples, sample_rate=sample_rate
    )
    assert_almost_equal(
        forward_shifted_samples, np.array([0.25, 0.125, 1.0, 0.5], dtype=np.float32)
    )
    assert forward_shifted_samples.dtype == np.float32
    assert len(forward_shifted_samples) == 4


def test_shift_without_rollover():
    samples = np.array([1.0, 0.5, 0.25, 0.125], dtype=np.float32)
    sample_rate = 16000

    forward_augmenter = Shift(
        min_shift=0.5, max_shift=0.5, rollover=False, fade_duration=0.0, p=1.0
    )
    forward_shifted_samples = forward_augmenter(
        samples=samples, sample_rate=sample_rate
    )
    assert_almost_equal(
        forward_shifted_samples, np.array([0.0, 0.0, 1.0, 0.5], dtype=np.float32)
    )
    assert forward_shifted_samples.dtype == np.float32
    assert len(forward_shifted_samples) == 4

    backward_augmenter = Shift(
        min_shift=-0.25, max_shift=-0.25, rollover=False, fade_duration=0.0, p=1.0
    )
    backward_shifted_samples = backward_augmenter(
        samples=samples, sample_rate=sample_rate
    )
    assert_almost_equal(
        backward_shifted_samples,
        np.array([0.5, 0.25, 0.125, 0.0], dtype=np.float32),
    )
    assert backward_shifted_samples.dtype == np.float32
    assert len(forward_shifted_samples) == 4


def test_shift_multichannel():
    samples = np.array(
        [[0.75, 0.5, -0.25, -0.125], [0.9, 0.5, -0.25, -0.125]], dtype=np.float32
    )
    sample_rate = 4000

    augment = Shift(min_shift=0.5, max_shift=0.5, fade_duration=0.0, p=1.0)
    processed_samples = augment(samples=samples, sample_rate=sample_rate)

    assert_almost_equal(
        processed_samples,
        np.array(
            [[-0.25, -0.125, 0.75, 0.5], [-0.25, -0.125, 0.9, 0.5]],
            dtype=np.float32,
        ),
    )
    assert processed_samples.dtype == np.float32


def test_shift_without_rollover_multichannel():
    samples = np.array(
        [[0.75, 0.5, -0.25, -0.125], [0.9, 0.5, -0.25, -0.125]], dtype=np.float32
    )
    sample_rate = 4000

    augment = Shift(
        min_shift=0.5, max_shift=0.5, rollover=False, fade_duration=0.0, p=1.0
    )
    processed_samples = augment(samples=samples, sample_rate=sample_rate)

    assert_almost_equal(
        processed_samples,
        np.array([[0.0, 0.0, 0.75, 0.5], [0.0, 0.0, 0.9, 0.5]], dtype=np.float32),
    )
    assert processed_samples.dtype == np.float32


def test_shift_fade():
    samples = np.array(
        [[1.0, 2.0, 3.0, 4.0, 5.0], [-1.0, -2.0, -3.0, -4.0, -5.0]],
        dtype=np.float32,
    )
    sample_rate = 4000

    augment = Shift(
        min_shift=0.5,
        max_shift=0.5,
        rollover=False,
        fade_duration=0.00075,  # 0.00075 * 4000 = 3
        p=1.0,
    )
    processed_samples = augment(samples=samples, sample_rate=sample_rate)

    assert processed_samples == pytest.approx(
        np.array(
            [[0.0, 0.0, 0.0, 1.4067, 3.0], [0.0, 0.0, 0.0, -1.4067, -3.0]],
            dtype=np.float32,
        ),
        abs=0.01,
    )


def test_shift_fade_rollover():
    samples = np.array(
        [[1.0, 2.0, 3.0, 4.0, 5.0], [-1.0, -2.0, -3.0, -4.0, -5.0]],
        dtype=np.float32,
    )
    sample_rate = 4000

    augment = Shift(
        min_shift=0.5,
        max_shift=0.5,
        rollover=True,
        fade_duration=0.00075,  # 0.00075 * 4000 = 3
        p=1.0,
    )
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert processed_samples == pytest.approx(
        np.array(
            [[2.81, 0.0, 0, 1.407, 3.0], [-2.81, 0.0, 0, -1.407, -3.0]],
            dtype=np.float32,
        ),
        abs=0.01,
    )


def test_shift_fade_rollover_2():
    samples = np.array(
        [[1.0, 2.0, 3.0, 4.0, 5.0], [-1.0, -2.0, -3.0, -4.0, -5.0]],
        dtype=np.float32,
    )
    sample_rate = 4000

    augment = Shift(
        min_shift=-0.5,
        max_shift=-0.5,
        rollover=True,
        fade_duration=0.00075,  # 0.00075 * 4000 = 3
        p=1.0,
    )
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert processed_samples == pytest.approx(
        np.array(
            [[3.0, 2.81, 0.0, 0.0, 1.407], [-3.0, -2.81, 0.0, -0.0, -1.407]],
            dtype=np.float32,
        ),
        abs=0.01,
    )


def test_shift_fade_rollover_3():
    samples = np.array(
        [[1.0, 2.0, 3.0, 4.0, 5.0], [-1.0, -2.0, -3.0, -4.0, -5.0]],
        dtype=np.float32,
    )
    sample_rate = 4000

    augment = Shift(
        min_shift=-0.5,
        max_shift=-0.5,
        rollover=True,
        fade_duration=1.0,
        p=1.0,
    )
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert processed_samples == pytest.approx(
        np.array(
            [
                [
                    3.0023373e-06,
                    1.0004364e-06,
                    0.0000000e00,
                    0.0000000e00,
                    5.0030241e-07,
                ],
                [
                    -3.0023373e-06,
                    -1.0004364e-06,
                    -0.0000000e00,
                    -0.0000000e00,
                    -5.0030241e-07,
                ],
            ],
            dtype=np.float32,
        ),
    )


def test_freeze_shift_with_different_sample_rates():
    samples1 = np.array([1.0, 0.5, 0.25, 0.125], dtype=np.float32)
    samples2 = np.array([1.0, 0.5, 0.25, 0.125, 0.3, 0.4, 0.5, 0.2], dtype=np.float32)
    sample_rate_1 = 1
    sample_rate_2 = 2

    augment = Shift(
        min_shift=0.5, max_shift=0.5, rollover=False, fade_duration=0.0, p=1.0
    )
    augment.randomize_parameters(samples1, sample_rate_1)
    augment.freeze_parameters()
    shifted_samples1 = augment(samples1, sample_rate_1)
    shifted_samples2 = augment(samples2, sample_rate_2)

    num_samples1_shifted = np.count_nonzero(shifted_samples1 == 0)
    num_samples2_shifted = np.count_nonzero(shifted_samples2 == 0)

    assert num_samples1_shifted == 2
    assert num_samples2_shifted == 4


def test_invalid_parameters():
    with pytest.raises(ValueError):
        Shift(min_shift=-1.5)
    with pytest.raises(ValueError):
        Shift(min_shift=1.5)
    with pytest.raises(ValueError):
        Shift(max_shift=-1.5)
    with pytest.raises(ValueError):
        Shift(max_shift=1.5)
    with pytest.raises(ValueError):
        Shift(min_shift=0.2, max_shift=0.1)
    with pytest.raises(ValueError):
        Shift(fade_duration=0.000001)
    with pytest.raises(ValueError):
        Shift(fade_duration=-1337)
    with pytest.raises(ValueError):
        Shift(shift_unit="lightyears")



import numpy as np

from audiomentations import Trim


def test_trim():
    sample_len = 1024
    samples1 = np.zeros((sample_len,), dtype=np.float32)
    samples2 = np.random.normal(0, 1, size=sample_len).astype(np.float32)
    sample_rate = 16000
    augmenter = Trim(top_db=20, p=1.0)
    samples_in = np.hstack((samples1, samples2))
    assert len(samples_in) == sample_len * 2
    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)

    assert samples_out.dtype == np.float32
    assert len(samples_out) < sample_len * 2



from typing import Optional, Union, Sequence

import numpy as np
from numpy.typing import NDArray


class WeightedChoiceSampler:
    """
    Draw integer indices according to a (possibly weighted) discrete distribution.

    Parameters
    ----------
    weights : Sequence[float] | NDArray | None, default None
        Normalised or un-normalised, non-negative weights for each outcome.
        If *None*, a uniform distribution over `num_items` is used.
    num_items : int | None, default None
        Number of outcomes when `weights is None`. Ignored otherwise.

    Examples
    --------
    >>> # weighted
    >>> s = WeightedChoiceSampler(weights=[0.2, 0.8])
    >>> s.sample(size=5)
    array([1, 0, 1, 1, 1])

    >>> # uniformly among 10 choices
    >>> s = WeightedChoiceSampler(num_items=10)     # weights=None implied
    >>> s.sample(size=5)
    array([7, 2, 3, 0, 9])
    """

    def __init__(
        self,
        weights: Optional[Union[Sequence[float], NDArray]] = None,
        *,
        num_items: Optional[int] = None,
    ):
        if weights is None:
            if num_items is None or num_items <= 0:
                raise ValueError(
                    "If 'weights' is None you must supply a positive 'num_items'."
                )
            self._uniform = True
            self.num_items = int(num_items)
            self.cdf = None
        else:
            weights = np.asarray(weights, dtype=float)
            if weights.ndim != 1:
                raise ValueError("weights must be one-dimensional")
            if np.any(weights < 0):
                raise ValueError("weights must be non-negative")
            total = float(weights.sum())
            if total == 0:
                raise ValueError("Sum of weights must be > 0")

            weights /= total
            self.cdf = np.cumsum(weights)
            # Ensure the last element is exactly 1.0 due to potential floating point errors
            self.cdf[-1] = 1.0
            self._uniform = False
            self.num_items = weights.size

    def sample(self, size: int = 1) -> Union[NDArray[np.int32], NDArray[np.int64]]:
        """
        Draw `size` indices according to the stored weight distribution.
        """
        if size < 1:
            raise ValueError("size must be a positive int")

        if self._uniform:
            # O(1) per sample
            return np.random.randint(self.num_items, size=size)
        else:
            # O(log n) per sample
            random_vals = np.random.random(size)
            return np.searchsorted(self.cdf, random_vals, side="right")



import math
import os
import pytest
import warnings

import numpy as np

from audiomentations.core.audio_loading_utils import (
    load_sound_file,
)
from demo.demo import DEMO_DIR


def test_load_stereo_ogg_vorbis():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "background_noises", "hens.ogg"), sample_rate=None
    )
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    # Apparently, the exact duration may vary slightly based on which decoder is used
    assert samples.shape[0] >= 442575
    assert samples.shape[0] <= 443328

    max_value = np.amax(samples)
    assert max_value > 0.02
    assert max_value < 1.0

def test_load_mono_opus():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "bus.opus"), sample_rate=None
    )
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    # Apparently, the exact duration may vary slightly based on which decoder is used
    assert samples.shape[0] >= 36682
    assert samples.shape[0] <= 36994

    max_value = np.amax(samples)
    assert max_value > 0.3
    assert max_value < 1.0

def test_load_mono_m4a():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "testing.m4a"), sample_rate=None
    )
    assert sample_rate == 44100
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    max_value = np.amax(samples)
    assert max_value > 0.1
    assert max_value < 1.0

def test_load_mono_signed_16_bit_wav():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "acoustic_guitar_0.wav"), sample_rate=None
    )
    assert sample_rate == 16000
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    assert samples.shape[0] == 140544

    max_value = np.amax(samples)
    assert max_value > 0.5
    assert max_value < 1.0

def test_load_stereo_signed_16_bit_wav():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "stereo_16bit.wav"), sample_rate=None
    )
    assert sample_rate == 16000
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    assert samples.shape[0] == 17833

    max_value = np.amax(samples)
    assert max_value > 0.5
    assert max_value < 1.0

def test_load_mono_signed_24_bit_wav():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "signed_24bit.wav"), sample_rate=None
    )
    assert sample_rate == 48000
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    assert samples.shape[0] == 54514

    max_value = np.amax(samples)
    assert max_value > 0.09
    assert max_value < 1.0

def test_load_mono_signed_24_bit_wav2():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "mono_int24.wav"), sample_rate=None
    )
    assert sample_rate == 44100
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    assert samples.shape[-1] == 22

    max_value = np.amax(samples)
    assert max_value == pytest.approx(0.96750367)
    min_value = np.amin(samples)
    assert min_value == pytest.approx(-0.9822748)

def test_load_mono_signed_32_bit_wav():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "mono_int32.wav"), sample_rate=None
    )
    assert sample_rate == 44100
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    assert samples.shape[-1] == 22

    max_value = np.amax(samples)
    assert max_value == pytest.approx(0.96750367)
    min_value = np.amin(samples)
    assert min_value == pytest.approx(-0.9822748)

def test_load_mono_float64_wav():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "mono_float64.wav"), sample_rate=None
    )
    assert sample_rate == 44100
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    assert samples.shape[-1] == 22

    max_value = np.amax(samples)
    assert max_value == pytest.approx(0.96750367)
    min_value = np.amin(samples)
    assert min_value == pytest.approx(-0.9822748)

def test_load_stereo_signed_24_bit_wav():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "stereo_24bit.WAV"), sample_rate=None
    )
    assert sample_rate == 16000
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    assert samples.shape[0] == 17833

    max_value = np.amax(samples)
    assert max_value > 0.5
    assert max_value < 1.0

def test_load_mono_ms_adpcm():
    samples, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "ms_adpcm.wav"), sample_rate=None
    )
    assert sample_rate == 11024
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    assert samples.shape[0] == 895500

    max_value = np.amax(samples)
    assert max_value > 0.3
    assert max_value < 1.0

def test_load_mono_ms_adpcm_and_resample():
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")

        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "ms_adpcm.wav"), sample_rate=16000
        )

        assert len(w) >= 1

    assert sample_rate == 16000
    assert samples.dtype == np.float32
    assert samples.ndim == 1

    assert samples.shape[0] == math.ceil(895500 * 16000 / 11024)

    max_value = np.amax(samples)
    assert max_value > 0.3
    assert max_value < 1.0



import json
import random

import fast_align_audio
import numpy as np
import pytest

from audiomentations import Limiter


@pytest.mark.parametrize(
    "samples_in",
    [
        np.random.normal(0, 1, size=1000).astype(np.float32),
        np.random.normal(0, 0.001, size=(1, 250)).astype(np.float32),
        np.random.normal(0, 0.1, size=(3, 8888)).astype(np.float32),
    ],
    ids=["normal_1d_1000", "normal_2d_1_250", "normal_2d_3_8888"],
)
def test_limiter(samples_in):
    augmenter = Limiter(p=1.0, min_attack=0.0025, max_attack=0.0025)
    std_in = np.mean(np.abs(samples_in))
    samples_out = augmenter(samples=samples_in, sample_rate=16000)
    std_out = np.mean(np.abs(samples_out))
    length = samples_in.shape[-1]

    samples_in_mono = samples_in
    samples_out_mono = samples_out
    if samples_in_mono.ndim > 1:
        samples_in_mono = samples_in_mono[0]
        samples_out_mono = samples_out_mono[0]
    offset, _ = fast_align_audio.find_best_alignment_offset(
        reference_signal=samples_in_mono,
        delayed_signal=samples_out_mono,
        max_offset_samples=length // 2,
        lookahead_samples=length // 2,
        method="corr",
    )
    # Check that the output is aligned with the input, i.e. no delay was introduced
    assert offset == 0

    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples_in.shape
    assert std_out < std_in


def test_limiter_validation():
    with pytest.raises(AssertionError):
        Limiter(min_attack=-0.5)


def test_serialize_parameters():
    random.seed(42)
    transform = Limiter(p=1.0)
    samples = np.random.normal(0, 1, size=1024).astype(np.float32)
    transform.randomize_parameters(samples, sample_rate=16000)
    json.dumps(transform.serialize_parameters())


def test_digital_silence():
    samples_in = np.zeros((1024,), np.float32)
    augmenter = Limiter(p=1.0)
    std_in = np.mean(np.abs(samples_in))
    samples_out = augmenter(samples=samples_in, sample_rate=16000)
    std_out = np.mean(np.abs(samples_out))
    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples_in.shape
    assert std_out == std_in == 0.0



import os
import random

import numpy as np
import pytest
from numpy.testing import assert_array_equal, assert_array_almost_equal

from audiomentations import (
    ClippingDistortion,
    AddBackgroundNoise,
    TimeMask,
    Shift,
    PolarityInversion,
    Gain,
    SomeOf,
)
from demo.demo import DEMO_DIR


def test_right_number_of_transforms_applied():
    random.seed(23)
    samples = np.array([0.25, 0.0, 0.1, -0.4], dtype=np.float32)
    sample_rate = 44100
    num_transforms_applied_list = []
    num_runs = 30
    list_transforms = [
        Gain(min_gain_db=-12, max_gain_db=-6, p=1.0),
        PolarityInversion(p=1.0),
    ]
    augmenter = SomeOf(1, list_transforms)

    for _ in range(num_runs):
        perturbed_samples = augmenter(samples=samples, sample_rate=sample_rate)
        was_gain_applied = 0.0 < abs(perturbed_samples[0]) < 0.25
        was_polarity_inversion_applied = perturbed_samples[0] < 0.0
        num_transforms_applied = sum(
            [
                1 if was_gain_applied else 0,
                1 if was_polarity_inversion_applied else 0,
            ]
        )
        num_transforms_applied_list.append(num_transforms_applied)
    assert np.mean(num_transforms_applied_list) == 1


def test_right_number_of_transforms_applied2():
    random.seed(23)
    samples = np.array([0.25, 0.0, 0.1, -0.4], dtype=np.float32)
    sample_rate = 44100
    num_transforms_applied_list = []
    num_runs = 30
    list_transforms = [
        Gain(min_gain_db=-12, max_gain_db=-6, p=1.0),
        PolarityInversion(p=1.0),
    ]
    augmenter = SomeOf((1, 2), list_transforms)

    for _ in range(num_runs):
        perturbed_samples = augmenter(samples=samples, sample_rate=sample_rate)
        was_gain_applied = 0.0 < abs(perturbed_samples[0]) < 0.25
        was_polarity_inversion_applied = perturbed_samples[0] < 0.0
        num_transforms_applied = sum(
            [
                1 if was_gain_applied else 0,
                1 if was_polarity_inversion_applied else 0,
            ]
        )
        num_transforms_applied_list.append(num_transforms_applied)
    assert np.mean(num_transforms_applied_list) == pytest.approx(1.5, abs=0.1)


def test_right_number_of_transforms_applied3():
    random.seed(23)
    samples = np.array([0.25, 0.0, 0.1, -0.4], dtype=np.float32)
    sample_rate = 44100
    num_transforms_applied_list = []
    num_runs = 30
    list_transforms = [
        Gain(min_gain_db=-12, max_gain_db=-6, p=1.0),
        PolarityInversion(p=1.0),
    ]
    augmenter = SomeOf((2, None), list_transforms)

    for _ in range(num_runs):
        perturbed_samples = augmenter(samples=samples, sample_rate=sample_rate)
        was_gain_applied = 0.0 < abs(perturbed_samples[0]) < 0.25
        was_polarity_inversion_applied = perturbed_samples[0] < 0.0
        num_transforms_applied = sum(
            [
                1 if was_gain_applied else 0,
                1 if was_polarity_inversion_applied else 0,
            ]
        )
        num_transforms_applied_list.append(num_transforms_applied)
    assert np.mean(num_transforms_applied_list) == 2


def test_right_number_of_transforms_applied4():
    random.seed(345)
    samples = np.array([0.25, 0.0, 0.1, -0.4], dtype=np.float32)
    sample_rate = 44100
    num_transforms_applied_list = []
    num_runs = 30
    list_transforms = [
        Gain(min_gain_db=-12, max_gain_db=-6, p=1.0),
        PolarityInversion(p=1.0),
    ]
    augmenter = SomeOf((0, None), list_transforms)

    for _ in range(num_runs):
        perturbed_samples = augmenter(samples=samples, sample_rate=sample_rate)
        was_gain_applied = 0.0 < abs(perturbed_samples[0]) < 0.25
        was_polarity_inversion_applied = perturbed_samples[0] < 0.0
        num_transforms_applied = sum(
            [
                1 if was_gain_applied else 0,
                1 if was_polarity_inversion_applied else 0,
            ]
        )
        num_transforms_applied_list.append(num_transforms_applied)
    assert np.mean(num_transforms_applied_list) == pytest.approx(1.0, abs=0.3)


def test_freeze_and_unfreeze_all_parameters():
    samples = np.array([0.25, 0.0, 0.1, -0.4], dtype=np.float32)
    sample_rate = 44100

    for _ in range(30):
        augmenter = SomeOf((1, None), [Gain(p=1.0), PolarityInversion(p=1.0)])
        perturbed_samples1 = augmenter(samples=samples, sample_rate=sample_rate)
        augmenter.freeze_parameters()
        for transform in augmenter.transforms:
            assert transform.are_parameters_frozen
        perturbed_samples2 = augmenter(samples=samples, sample_rate=sample_rate)

        assert_array_equal(perturbed_samples1, perturbed_samples2)

        augmenter.unfreeze_parameters()
        for transform in augmenter.transforms:
            assert not transform.are_parameters_frozen


def test_freeze_and_unfreeze_own_parameters():
    augmenter = SomeOf(
        (1, None),
        [
            Gain(p=1.0),
            PolarityInversion(p=1.0),
        ],
    )
    assert not augmenter.are_parameters_frozen
    for transform in augmenter.transforms:
        assert not transform.are_parameters_frozen

    augmenter.freeze_parameters(apply_to_children=False)
    assert augmenter.are_parameters_frozen
    for transform in augmenter.transforms:
        assert not transform.are_parameters_frozen

    augmenter.unfreeze_parameters(apply_to_children=False)
    assert not augmenter.are_parameters_frozen
    for transform in augmenter.transforms:
        assert not transform.are_parameters_frozen


def test_randomize_all_parameters_and_apply():
    samples = 1.0 / np.arange(1, 21, dtype=np.float32)
    sample_rate = 44100

    augmenter = SomeOf(
        (1, None),
        [
            AddBackgroundNoise(
                sounds_path=os.path.join(DEMO_DIR, "background_noises"),
                min_snr_db=15,
                max_snr_db=35,
                p=1.0,
            ),
            ClippingDistortion(p=1.0),
            TimeMask(min_band_part=0.2, max_band_part=0.5, p=1.0),
            Shift(min_shift=0.5, max_shift=0.5, p=1.0),
        ],
    )
    augmenter.freeze_parameters()
    augmenter.randomize_parameters(samples=samples, sample_rate=sample_rate)

    parameters = [transform.parameters for transform in augmenter.transforms]

    perturbed_samples1 = augmenter(samples=samples, sample_rate=sample_rate)
    perturbed_samples2 = augmenter(samples=samples, sample_rate=sample_rate)

    assert_array_equal(perturbed_samples1, perturbed_samples2)

    augmenter.unfreeze_parameters()

    for transform_parameters, transform in zip(parameters, augmenter.transforms):
        assert transform_parameters == transform.parameters
        assert not transform.are_parameters_frozen


def test_randomize_only_own_parameters():
    samples = 1.0 / np.arange(1, 21, dtype=np.float32)
    sample_rate = 44100

    augmenter = SomeOf(
        (1, None),
        [
            Gain(p=1.0),
            PolarityInversion(p=1.0),
        ],
    )
    augmenter.randomize_parameters(samples, sample_rate, apply_to_children=True)
    own_parameters_before = (augmenter.should_apply, augmenter.transform_indexes)

    parameters = [transform.parameters for transform in augmenter.transforms]

    own_parameters_changed = False
    for i in range(30):
        augmenter.randomize_parameters(samples, sample_rate, apply_to_children=False)
        if (
            augmenter.should_apply,
            augmenter.transform_indexes,
        ) != own_parameters_before:
            own_parameters_changed = True
            break
    assert own_parameters_changed

    # Check that the children's parameters are still the same
    for transform_parameters, transform in zip(parameters, augmenter.transforms):
        assert transform_parameters == transform.parameters
        assert not transform.are_parameters_frozen



import os
import random
import tempfile
import uuid
from typing import Literal

import librosa
import numpy as np
import sys
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import (
    convert_float_samples_to_int16,
    get_max_abs_amplitude,
)


class Mp3Compression(BaseWaveformTransform):
    """Compress the audio using an MP3 encoder to lower the audio quality.
    This may help machine learning models deal with compressed, low-quality audio.

    This transform depends on either lameenc or pydub/ffmpeg.

    Note that bitrates below 32 kbps are only supported for low sample rates (up to 24000 Hz).

    Note: When using the lameenc backend, the output may be slightly longer than the input due
    to the fact that the LAME encoder inserts some silence at the beginning of the audio.

    Warning: This transform writes to disk, so it may be slow. Ideally, the work should be done
    in memory. Contributions are welcome.
    """

    supports_multichannel = True

    SUPPORTED_BITRATES = [
        8,
        16,
        24,
        32,
        40,
        48,
        56,
        64,
        80,
        96,
        112,
        128,
        144,
        160,
        192,
        224,
        256,
        320,
    ]

    def __init__(
        self,
        min_bitrate: int = 8,
        max_bitrate: int = 64,
        backend: Literal["pydub", "lameenc"] = "pydub",
        p: float = 0.5,
    ):
        """
        :param min_bitrate: Minimum bitrate in kbps
        :param max_bitrate: Maximum bitrate in kbps
        :param backend: "pydub" or "lameenc".
            Pydub may use ffmpeg under the hood.
                Pros: Seems to avoid introducing latency in the output.
                Cons: Slower than lameenc.
            lameenc:
                Pros: You can set the quality parameter in addition to bitrate.
                Cons: Seems to introduce some silence at the start of the audio.
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        if min_bitrate < self.SUPPORTED_BITRATES[0]:
            raise ValueError(
                "min_bitrate must be greater than or equal to"
                f" {self.SUPPORTED_BITRATES[0]}"
            )
        if max_bitrate > self.SUPPORTED_BITRATES[-1]:
            raise ValueError(
                "max_bitrate must be less than or equal to"
                f" {self.SUPPORTED_BITRATES[-1]}"
            )
        if max_bitrate < min_bitrate:
            raise ValueError("max_bitrate must be >= min_bitrate")

        is_any_supported_bitrate_in_range = any(
            min_bitrate <= bitrate <= max_bitrate for bitrate in self.SUPPORTED_BITRATES
        )
        if not is_any_supported_bitrate_in_range:
            raise ValueError(
                "There is no supported bitrate in the range between the specified"
                " min_bitrate and max_bitrate. The supported bitrates are:"
                f" {self.SUPPORTED_BITRATES}"
            )

        self.min_bitrate = min_bitrate
        self.max_bitrate = max_bitrate
        if backend not in ("pydub", "lameenc"):
            raise ValueError('backend must be set to either "pydub" or "lameenc"')
        self.backend = backend
        self.post_gain_factor = None

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            bitrate_choices = [
                bitrate
                for bitrate in self.SUPPORTED_BITRATES
                if self.min_bitrate <= bitrate <= self.max_bitrate
            ]
            self.parameters["bitrate"] = random.choice(bitrate_choices)

    def apply(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        if self.backend == "lameenc":
            return self.apply_lameenc(samples, sample_rate)
        elif self.backend == "pydub":
            return self.apply_pydub(samples, sample_rate)
        else:
            raise Exception("Backend {} not recognized".format(self.backend))

    def maybe_pre_gain(self, samples):
        """
        If the audio is too loud, gain it down to avoid distortion in the audio file to
        be encoded.
        """
        greatest_abs_sample = get_max_abs_amplitude(samples)
        if greatest_abs_sample > 1.0:
            self.post_gain_factor = greatest_abs_sample
            samples = samples * (1.0 / greatest_abs_sample)
        else:
            self.post_gain_factor = None
        return samples

    def maybe_post_gain(self, samples):
        """If the audio was pre-gained down earlier, post-gain it up to compensate here."""
        if self.post_gain_factor is not None:
            samples = samples * self.post_gain_factor
        return samples

    def apply_lameenc(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        try:
            import lameenc
        except ImportError:
            print(
                (
                    "Failed to import the lame encoder. Maybe it is not installed? "
                    "To install the optional lameenc dependency of audiomentations,"
                    " do `pip install audiomentations[extras]` or simply"
                    " `pip install lameenc`"
                ),
                file=sys.stderr,
            )
            raise

        assert samples.dtype == np.float32

        samples = self.maybe_pre_gain(samples)

        int_samples = convert_float_samples_to_int16(samples).T

        num_channels = 1 if samples.ndim == 1 else samples.shape[0]

        encoder = lameenc.Encoder()
        encoder.set_bit_rate(self.parameters["bitrate"])
        encoder.set_in_sample_rate(sample_rate)
        encoder.set_channels(num_channels)
        encoder.set_quality(7)  # 2 = highest, 7 = fastest
        encoder.silence()

        mp3_data = encoder.encode(int_samples.tobytes())
        mp3_data += encoder.flush()

        # Write a temporary MP3 file that will then be decoded
        tmp_dir = tempfile.gettempdir()
        tmp_file_path = os.path.join(
            tmp_dir, "tmp_compressed_{}.mp3".format(str(uuid.uuid4())[0:12])
        )
        with open(tmp_file_path, "wb") as f:
            f.write(mp3_data)

        degraded_samples, _ = librosa.load(tmp_file_path, sr=sample_rate, mono=False)

        os.unlink(tmp_file_path)

        degraded_samples = self.maybe_post_gain(degraded_samples)

        if num_channels == 1:
            if int_samples.ndim == 1 and degraded_samples.ndim == 2:
                degraded_samples = np.ravel(degraded_samples)
            elif int_samples.ndim == 2 and degraded_samples.ndim == 1:
                degraded_samples = degraded_samples.reshape((1, -1))

        return degraded_samples

    def apply_pydub(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        try:
            import pydub
        except ImportError:
            print(
                (
                    "Failed to import pydub. Maybe it is not installed? "
                    "To install the optional pydub dependency of audiomentations,"
                    " do `pip install audiomentations[extras]` or simply"
                    " `pip install pydub`"
                ),
                file=sys.stderr,
            )
            raise

        assert samples.dtype == np.float32

        samples = self.maybe_pre_gain(samples)

        int_samples = convert_float_samples_to_int16(samples).T
        num_channels = 1 if samples.ndim == 1 else samples.shape[0]
        audio_segment = pydub.AudioSegment(
            int_samples.tobytes(),
            frame_rate=sample_rate,
            sample_width=int_samples.dtype.itemsize,
            channels=num_channels,
        )

        tmp_dir = tempfile.gettempdir()
        tmp_file_path = os.path.join(
            tmp_dir, "tmp_compressed_{}.mp3".format(str(uuid.uuid4())[0:12])
        )

        bitrate_string = "{}k".format(self.parameters["bitrate"])
        file_handle = audio_segment.export(tmp_file_path, bitrate=bitrate_string)
        file_handle.close()

        degraded_samples, _ = librosa.load(tmp_file_path, sr=sample_rate, mono=False)

        os.unlink(tmp_file_path)

        degraded_samples = self.maybe_post_gain(degraded_samples)

        if num_channels == 1:
            if int_samples.ndim == 1 and degraded_samples.ndim == 2:
                degraded_samples = np.ravel(degraded_samples)
            elif int_samples.ndim == 2 and degraded_samples.ndim == 1:
                degraded_samples = degraded_samples.reshape((1, -1))

        return degraded_samples



import json
import os
import pickle

import numpy as np
import pytest

from audiomentations import (
    AddBackgroundNoise,
    AddColorNoise,
    AddGaussianNoise,
    AddGaussianSNR,
    AddShortNoises,
    AdjustDuration,
    AirAbsorption,
    Aliasing,
    ApplyImpulseResponse,
    BandPassFilter,
    BandStopFilter,
    BitCrush,
    Clip,
    ClippingDistortion,
    Gain,
    GainTransition,
    HighPassFilter,
    HighShelfFilter,
    Lambda,
    Limiter,
    LoudnessNormalization,
    LowPassFilter,
    LowShelfFilter,
    Mp3Compression,
    Normalize,
    Padding,
    PeakingFilter,
    PitchShift,
    PolarityInversion,
    RepeatPart,
    Resample,
    Reverse,
    RoomSimulator,
    SevenBandParametricEQ,
    Shift,
    TanhDistortion,
    TimeMask,
    TimeStretch,
    Trim,
)
from demo.demo import DEMO_DIR


@pytest.mark.parametrize(
    "transform",
    [
        AddBackgroundNoise(
            sounds_path=os.path.join(DEMO_DIR, "background_noises"), p=1.0
        ),
        AddColorNoise(p=1.0),
        AddGaussianNoise(p=1.0),
        AddGaussianSNR(p=1.0),
        AddShortNoises(sounds_path=os.path.join(DEMO_DIR, "background_noises"), p=1.0),
        AdjustDuration(duration_samples=500, p=1.0),
        AirAbsorption(p=1.0),
        Aliasing(p=1.0),
        ApplyImpulseResponse(ir_path=os.path.join(DEMO_DIR, "ir"), p=1.0),
        BandPassFilter(p=1.0),
        BandStopFilter(p=1.0),
        BitCrush(p=1.0),
        Clip(p=1.0),
        ClippingDistortion(p=1.0),
        Gain(p=1.0),
        GainTransition(p=1.0),
        HighPassFilter(p=1.0),
        HighShelfFilter(p=1.0),
        Lambda(transform=Gain(p=1.0), p=1.0),
        Limiter(p=1.0),
        LoudnessNormalization(p=1.0),
        LowPassFilter(p=1.0),
        LowShelfFilter(p=1.0),
        Mp3Compression(p=1.0),
        Normalize(p=1.0),
        Padding(p=1.0),
        PeakingFilter(p=1.0),
        PitchShift(p=1.0),
        PolarityInversion(p=1.0),
        RepeatPart(p=1.0),
        Resample(p=1.0),
        Reverse(p=1.0),
        # RoomSimulator(p=1.0),  # TODO: Fix TypeError: cannot pickle 'pyroomacoustics.libroom.Wall' object
        SevenBandParametricEQ(p=1.0),
        Shift(p=1.0),
        TanhDistortion(p=1.0),
        TimeMask(p=1.0),
        TimeStretch(p=1.0),
        Trim(p=1.0),
    ],
    ids=lambda transform: transform.__class__.__name__,
)
@pytest.mark.parametrize(
    "samples", [np.random.normal(0, 1, size=17640).astype(np.float32)]
)
def test_param_json_serialization_and_transform_picklability(transform, samples):
    transform.randomize_parameters(samples, sample_rate=44100)
    json.dumps(transform.serialize_parameters())

    # Picklability is required for multiprocessing compatibility
    pickled = pickle.dumps(transform)
    pickle.loads(pickled)



import random

import numpy as np
from numpy.typing import NDArray
from scipy.signal import sosfilt, sosfilt_zi

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import (
    convert_frequency_to_mel,
    convert_mel_to_frequency,
)


class HighShelfFilter(BaseWaveformTransform):
    """
    A high shelf filter is a filter that either boosts (increases amplitude) or cuts
    (decreases amplitude) frequencies above a certain center frequency. This transform
    applies a high-shelf filter at a specific center frequency in hertz.
    The gain at Nyquist frequency is controlled by `{min,max}_gain_db` (note: can be positive or negative!).
    Filter coefficients are taken from the W3 Audio EQ Cookbook: https://www.w3.org/TR/audio-eq-cookbook/
    """

    supports_multichannel = True

    def __init__(
        self,
        min_center_freq: float = 300.0,
        max_center_freq: float = 7500.0,
        min_gain_db: float = -18.0,
        max_gain_db: float = 18.0,
        min_q: float = 0.1,
        max_q: float = 0.999,
        p: float = 0.5,
    ):
        """
        :param min_center_freq: The minimum center frequency of the shelving filter
        :param max_center_freq: The maximum center frequency of the shelving filter
        :param min_gain_db: The minimum gain at the Nyquist frequency
        :param max_gain_db: The maximum gain at the Nyquist frequency
        :param min_q: The minimum quality factor Q. The higher the Q, the steeper the
            transition band will be.
        :param max_q: The maximum quality factor Q. The higher the Q, the steeper the
            transition band will be.
        :param p: The probability of applying this transform
        """

        assert (
            min_center_freq <= max_center_freq
        ), "`min_center_freq` should be no greater than `max_center_freq`"
        assert (
            min_gain_db <= max_gain_db
        ), "`min_gain_db` should be no greater than `max_gain_db`"

        assert 0 < min_q <= 1, "`min_q` should be greater than 0 and less or equal to 1"
        assert 0 < max_q <= 1, "`max_q` should be greater than 0 and less or equal to 1"

        super().__init__(p)

        self.min_center_freq = min_center_freq
        self.max_center_freq = max_center_freq

        self.min_gain_db = min_gain_db
        self.max_gain_db = max_gain_db

        self.min_q = min_q
        self.max_q = max_q

    def _get_biquad_coefficients_from_input_parameters(
        self, center_freq, gain_db, q_factor, sample_rate
    ):
        normalized_frequency = 2 * np.pi * center_freq / sample_rate
        gain = 10 ** (gain_db / 40)
        alpha = np.sin(normalized_frequency) / 2 / q_factor

        b0 = gain * (
            (gain + 1)
            + (gain - 1) * np.cos(normalized_frequency)
            + 2 * np.sqrt(gain) * alpha
        )

        b1 = -2 * gain * ((gain - 1) + (gain + 1) * np.cos(normalized_frequency))

        b2 = gain * (
            (gain + 1)
            + (gain - 1) * np.cos(normalized_frequency)
            - 2 * np.sqrt(gain) * alpha
        )

        a0 = (
            (gain + 1)
            - (gain - 1) * np.cos(normalized_frequency)
            + 2 * np.sqrt(gain) * alpha
        )

        a1 = 2 * ((gain - 1) - (gain + 1) * np.cos(normalized_frequency))

        a2 = (
            (gain + 1)
            - (gain - 1) * np.cos(normalized_frequency)
            - 2 * np.sqrt(gain) * alpha
        )

        # Return it in `sos` format
        sos = np.array([[b0 / a0, b1 / a0, b2 / a0, 1, a1 / a0, a2 / a0]])

        return sos

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)

        center_mel = np.random.uniform(
            low=convert_frequency_to_mel(self.min_center_freq),
            high=convert_frequency_to_mel(self.max_center_freq),
        )
        self.parameters["center_freq"] = convert_mel_to_frequency(center_mel)
        self.parameters["gain_db"] = random.uniform(self.min_gain_db, self.max_gain_db)
        self.parameters["q_factor"] = random.uniform(self.min_q, self.max_q)

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        nyquist_freq = sample_rate // 2
        center_freq = self.parameters["center_freq"]
        if center_freq > nyquist_freq:
            # Ensure that the center frequency is below the Nyquist
            # frequency to avoid filter instability
            center_freq = nyquist_freq * 0.9999

        sos = self._get_biquad_coefficients_from_input_parameters(
            center_freq,
            self.parameters["gain_db"],
            self.parameters["q_factor"],
            sample_rate,
        )

        # The processing takes place here
        zi = sosfilt_zi(sos)
        if len(samples.shape) == 1:
            processed_samples, _ = sosfilt(sos, samples, zi=zi * samples[0])
            processed_samples = processed_samples.astype(np.float32)
        else:
            processed_samples = np.zeros_like(samples, dtype=np.float32)
            for chn_idx in range(samples.shape[0]):
                processed_samples[chn_idx, :], _ = sosfilt(
                    sos, samples[chn_idx, :], zi=zi * samples[chn_idx, 0]
                )

        return processed_samples



import librosa
import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class Trim(BaseWaveformTransform):
    """
    Trim leading and trailing silence from an audio signal using librosa.effects.trim
    """

    supports_multichannel = True

    def __init__(self, top_db: float = 30.0, p: float = 0.5):
        """
        :param top_db: The threshold (in Decibels) below reference to consider as silence
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        self.top_db = top_db

    def apply(self, samples: NDArray[np.float32], sample_rate: int):
        samples, _ = librosa.effects.trim(samples, top_db=self.top_db)
        return samples



import os

import fast_align_audio
import numpy as np
import pytest

from audiomentations import Mp3Compression
from audiomentations.core.audio_loading_utils import load_sound_file
from demo.demo import DEMO_DIR


def mse(signal1, signal2):
    """Mean squared error"""
    return np.mean((signal1 - signal2) ** 2)


@pytest.mark.parametrize(
    "backend",
    ["pydub", "lameenc"],
)
@pytest.mark.parametrize(
    "shape",
    [(44100,), (1, 22049), (2, 10000)],
)
def test_apply_mp3_compression(backend: str, shape: tuple):
    samples_in = np.random.normal(0, 1, size=shape).astype(np.float32)
    sample_rate = 44100
    augmenter = Mp3Compression(p=1.0, min_bitrate=48, max_bitrate=48, backend=backend)

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert len(shape) == len(samples_out.shape)
    assert samples_out.dtype == np.float32
    assert samples_out.shape[-1] >= shape[-1]
    assert samples_out.shape[-1] < shape[-1] + 3000
    if len(shape) == 2:
        assert samples_out.shape[0] == shape[0]


@pytest.mark.parametrize(
    "backend",
    ["pydub", "lameenc"],
)
@pytest.mark.parametrize(
    "shape",
    [(16000,), (1, 12049), (2, 5000)],
)
def test_apply_mp3_compression_low_bitrate(backend: str, shape: tuple):
    samples_in = np.random.normal(0, 1, size=shape).astype(np.float32)
    sample_rate = 16000
    augmenter = Mp3Compression(p=1.0, min_bitrate=8, max_bitrate=8, backend=backend)

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert len(shape) == len(samples_out.shape)
    assert samples_out.dtype == np.float32
    assert samples_out.shape[-1] >= shape[-1]
    assert samples_out.shape[-1] < shape[-1] + 3100
    if len(shape) == 2:
        assert samples_out.shape[0] == shape[0]


def test_invalid_argument_combination():
    with pytest.raises(ValueError):
        _ = Mp3Compression(min_bitrate=400, max_bitrate=800)

    with pytest.raises(ValueError):
        _ = Mp3Compression(min_bitrate=2, max_bitrate=128)

    with pytest.raises(ValueError):
        _ = Mp3Compression(min_bitrate=64, max_bitrate=32)

    with pytest.raises(ValueError):
        _ = Mp3Compression(min_bitrate=66, max_bitrate=67)

    with pytest.raises(ValueError):
        _ = Mp3Compression(backend="both")


def test_too_loud_input():
    """Check that we avoid wrap distortion if input is too loud"""
    samples_in, sample_rate = load_sound_file(
        os.path.join(DEMO_DIR, "perfect-alley1.ogg"),
        sample_rate=None,
        mono=True,
    )
    samples_in = samples_in[..., 0:196000]
    samples_in *= 10
    augmenter = Mp3Compression(
        min_bitrate=320, max_bitrate=320, backend="lameenc", p=1.0
    )

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    samples_out = samples_out[..., 0 : samples_in.shape[-1]]
    assert samples_out.dtype == np.float32

    # Apply delay compensation, so we can compare output with input sample by sample
    offset, _ = fast_align_audio.find_best_alignment_offset(
        reference_signal=samples_in,
        delayed_signal=samples_out,
        max_offset_samples=3100,
        lookahead_samples=10_000,
        method="corr",
    )

    snippet_length = 42000
    delay_compensated_samples_out = samples_out[..., offset : offset + snippet_length]
    samples_in = samples_in[..., :snippet_length]
    mse_value = mse(samples_in, delay_compensated_samples_out)

    assert mse_value < 0.00001



import numpy as np
import pytest
import scipy
import scipy.signal

from audiomentations import LowPassFilter
from tests.utils import get_chirp_test, get_randn_test

DEBUG = False


@pytest.mark.parametrize("cutoff_frequency", [1000])
@pytest.mark.parametrize("rolloff", [6, 24])
@pytest.mark.parametrize("zero_phase", [False, True])
def test_one_single_input(
    cutoff_frequency,
    rolloff,
    zero_phase,
):
    sample_rate = 8000

    # Parameters for computing periodograms.
    # When examining lower frequencies we need to have
    # a high nfft number.

    nfft = 2048 * 2
    nperseg = 128

    samples = get_chirp_test(sample_rate, 10)

    # Expected dB drop at fc
    if zero_phase:
        expected_db_drop = 6
    else:
        expected_db_drop = 3

    if zero_phase and rolloff % 12 != 0:
        with pytest.raises(AssertionError):
            augment = LowPassFilter(
                min_cutoff_freq=cutoff_frequency,
                max_cutoff_freq=cutoff_frequency,
                min_rolloff=rolloff,
                max_rolloff=rolloff,
                zero_phase=zero_phase,
                p=1.0,
            )
        return
    else:
        augment = LowPassFilter(
            min_cutoff_freq=cutoff_frequency,
            max_cutoff_freq=cutoff_frequency,
            min_rolloff=rolloff,
            max_rolloff=rolloff,
            zero_phase=zero_phase,
            p=1.0,
        )

    processed_samples = augment(samples=samples, sample_rate=sample_rate)

    # Compute periodograms
    wx, samples_pxx = scipy.signal.welch(
        samples,
        fs=sample_rate,
        nfft=nfft,
        nperseg=nperseg,
        scaling="spectrum",
        window="hann",
    )
    _, processed_samples_pxx = scipy.signal.welch(
        processed_samples,
        fs=sample_rate,
        nperseg=nperseg,
        nfft=nfft,
        scaling="spectrum",
        window="hann",
    )

    if DEBUG:
        import matplotlib.pyplot as plt

        plt.title(
            f"Filter type: High Roll-off:{rolloff}dB/octave Zero-phase:{zero_phase}"
        )
        plt.plot(wx, 10 * np.log10(np.abs(samples_pxx)))
        plt.plot(wx, 10 * np.log10(np.abs(processed_samples_pxx)), ":")
        plt.legend(["Input signal", f"Lowpassed at f_c={cutoff_frequency:.2f}"])
        plt.axvline(cutoff_frequency, color="red", linestyle=":")
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Magnitude (dB)")
        plt.show()

    assert processed_samples.shape == samples.shape
    assert processed_samples.dtype == np.float32

    frequencies_of_interest = np.array([0, cutoff_frequency, cutoff_frequency * 2])
    expected_differences = np.array([0.0, expected_db_drop, expected_db_drop + rolloff])

    # Tolerances for the differences in dB
    tolerances = np.array([1.0, 2.0, 5.0])

    for n, freq in enumerate(frequencies_of_interest):
        input_value_db = 10 * np.log10(
            samples_pxx[int(np.round(nfft / sample_rate * freq))]
        )

        output_value_db = 10 * np.log10(
            processed_samples_pxx[int(np.round(nfft / sample_rate * freq))]
        )

        assert np.isclose(
            input_value_db - output_value_db,
            expected_differences[n],
            atol=tolerances[n],
        )


@pytest.mark.parametrize("samples", [get_chirp_test(8000, 40)])
@pytest.mark.parametrize("rolloff", [12, 120])
@pytest.mark.parametrize("zero_phase", [False, True])
@pytest.mark.parametrize("num_channels", [1, 2, 3])
def test_n_channel_input(samples, rolloff, zero_phase, num_channels):
    sample_rate = 8000
    samples = get_randn_test(sample_rate, 10)

    # Convert to 2D N channels
    n_channels = np.tile(samples, (num_channels, 1))

    augment = LowPassFilter(
        min_cutoff_freq=1000,
        max_cutoff_freq=1000,
        min_rolloff=rolloff,
        max_rolloff=rolloff,
        zero_phase=zero_phase,
        p=1.0,
    )

    processed_samples = augment(samples=samples, sample_rate=sample_rate)

    processed_n_channels = augment(samples=n_channels, sample_rate=sample_rate)

    assert processed_n_channels.shape[0] == num_channels
    assert processed_n_channels.shape == n_channels.shape
    assert processed_n_channels.dtype == np.float32

    # Check that the processed 2D channel version applies the same effect
    # as the passband version.
    for _, channel in enumerate(processed_n_channels):
        if DEBUG:
            import matplotlib.pyplot as plt

            plt.title(
                f"Filter type: Low-pass Roll-off:{rolloff}dB/octave Zero-phase:{zero_phase}"
            )
            plt.plot(processed_samples)
            plt.plot(channel, "r--")

            plt.legend(["1D", "2D"])
            plt.show()
        assert np.allclose(channel, processed_samples)



import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class Clip(BaseWaveformTransform):
    """
    Clip audio by specified values. e.g. set a_min=-1.0 and a_max=1.0 to ensure that no
    samples in the audio exceed that extent. This can be relevant for avoiding integer
    overflow or underflow (which results in unintended wrap distortion that can sound
    horrible) when exporting to e.g. 16-bit PCM wav.

    Another way of ensuring that all values stay between -1.0 and 1.0 is to apply
    PeakNormalization.

    This transform is different from ClippingDistortion in that it takes fixed values
    for clipping instead of clipping a random percentile of the samples. Arguably, this
    transform is not very useful for data augmentation. Instead, think of it as a very
    cheap and harsh limiter (for samples that exceed the allotted extent) that can
    sometimes be useful at the end of a data augmentation pipeline.
    """

    supports_multichannel = True

    def __init__(self, a_min: float = -1.0, a_max: float = 1.0, p: float = 0.5):
        """
        :param a_min: float, minimum value for clipping
        :param a_max: float, maximum value for clipping
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        assert a_min < a_max
        self.a_min = a_min
        self.a_max = a_max

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        return np.clip(samples, self.a_min, self.a_max)



import random

import numpy as np
import pytest

from audiomentations import GainTransition


@pytest.mark.parametrize(
    "samples",
    [
        # Test both mono and stereo
        np.random.uniform(low=-0.5, high=0.5, size=(1234,)).astype(np.float32),
        np.random.uniform(low=-0.5, high=0.5, size=(2, 5678)).astype(np.float32),
    ],
)
def test_gain_transition_fraction(samples):
    np.random.seed(42)
    random.seed(42)
    sample_rate = 8000

    augment = GainTransition(p=1.0)
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert not np.allclose(samples, processed_samples)
    assert processed_samples.shape == samples.shape
    assert processed_samples.dtype == np.float32


def test_gain_transition_seconds():
    np.random.seed(40)
    random.seed(40)
    samples = np.random.uniform(low=-0.5, high=0.5, size=(2345,)).astype(np.float32)
    sample_rate = 16000

    augment = GainTransition(
        min_duration=0.2, max_duration=0.3, duration_unit="seconds", p=1.0
    )
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert not np.allclose(samples, processed_samples)
    assert processed_samples.shape == samples.shape
    assert processed_samples.dtype == np.float32


def test_gain_transition_samples():
    np.random.seed(1337)
    random.seed(1337)
    samples = np.random.uniform(low=-0.5, high=0.5, size=(3456,)).astype(np.float32)
    sample_rate = 32000

    augment = GainTransition(
        min_duration=1, max_duration=5000, duration_unit="samples", p=1.0
    )
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert not np.allclose(samples, processed_samples)
    assert processed_samples.shape == samples.shape
    assert processed_samples.dtype == np.float32


def test_invalid_params():
    with pytest.raises(ValueError):
        GainTransition(
            min_duration=-12, max_duration=324, duration_unit="samples", p=1.0
        )

    with pytest.raises(ValueError):
        GainTransition(min_duration=45, max_duration=10, duration_unit="samples", p=1.0)

    with pytest.raises(ValueError):
        GainTransition(min_gain_db=6.0, max_gain_db=-24.0, p=1.0)

    augment = GainTransition(
        min_duration=45, max_duration=45, duration_unit="lightyears", p=1.0
    )
    with pytest.raises(ValueError):
        augment(np.zeros(40, dtype=np.float32), 16000)



import numpy as np
import pytest

from audiomentations import TimeStretch


@pytest.mark.parametrize("method", ["librosa_phase_vocoder", "signalsmith_stretch"])
def test_dynamic_length(method):
    samples = np.zeros((2048,), dtype=np.float32)
    sample_rate = 16000
    augmenter = TimeStretch(
        min_rate=0.8, max_rate=0.9, leave_length_unchanged=False, method=method, p=1.0
    )

    samples_out = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples_out.dtype == np.float32
    assert samples_out.ndim == samples.ndim
    assert samples_out.shape[-1] > 2048


@pytest.mark.parametrize("method", ["librosa_phase_vocoder", "signalsmith_stretch"])
def test_fixed_length(method):
    samples = np.zeros((2048,), dtype=np.float32)
    sample_rate = 16000
    augmenter = TimeStretch(
        min_rate=0.8, max_rate=0.9, leave_length_unchanged=True, method=method, p=1.0
    )

    samples_out = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples_out.dtype == np.float32
    assert samples_out.ndim == samples.ndim
    assert samples_out.shape[-1] == 2048


@pytest.mark.parametrize("method", ["librosa_phase_vocoder", "signalsmith_stretch"])
def test_multichannel(method):
    num_channels = 3
    samples = np.random.normal(0, 0.1, size=(num_channels, 5555)).astype(np.float32)
    sample_rate = 16000
    augmenter = TimeStretch(
        min_rate=0.8, max_rate=0.9, leave_length_unchanged=True, method=method, p=1.0
    )

    samples_out = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == samples_out.dtype
    assert samples.shape == samples_out.shape
    for i in range(num_channels):
        assert not np.allclose(samples[i], samples_out[i])


def test_invalid_params():
    with pytest.raises(ValueError):
        TimeStretch(min_rate=0.0)
    with pytest.raises(ValueError):
        TimeStretch(max_rate=11.0)
    with pytest.raises(ValueError):
        TimeStretch(min_rate=2.0, max_rate=1.0)
    with pytest.raises(ValueError):
        TimeStretch(min_rate=1.0, max_rate=2.0, method="invalid")



import random

import numpy as np
import pytest

from audiomentations import RoomSimulator

DEBUG = False


def get_sinc_impulse(sample_rate, duration):
    """Create a `duration` seconds chirp from 0Hz to `Nyquist frequency`"""
    n = np.arange(-duration / 2, duration / 2, 1 / sample_rate)

    # Full band sinc impulse centered at half the duration
    samples = 2 * 0.25 * np.sinc(2 * sample_rate / 4 * n)

    return samples.astype(np.float32)


def test_simulate_apply_parity():
    """
    Tests whether RoomSimulator.apply gives the same result as Roomsimulator.room.simulate() in the 1D case.

    This mainly tests that we took into consideration and compensated about the delays introduced when pyroomacoustics
    computes the room impulse response.

    See:[Create the Room Impulse Response](https://pyroomacoustics.readthedocs.io/en/pypi-release/pyroomacoustics.room.html?highlight=simulate#)
    """

    random.seed(1)
    sample_rate = 16000
    samples = get_sinc_impulse(sample_rate, 10)

    augment = RoomSimulator()
    augmented_samples_apply = augment(samples=samples, sample_rate=sample_rate)
    augment.room.simulate()
    augmented_samples_simulate = augment.room.mic_array.signals.astype(
        np.float32
    ).flatten()

    assert augmented_samples_simulate == pytest.approx(augmented_samples_apply)


def test_failing_case():
    """Failed case which identified a bug where the room created was not rectangular"""
    sample_rate = 16000

    samples = get_sinc_impulse(sample_rate, 10)

    augment = RoomSimulator(
        min_size_x=3.0,
        min_size_y=4.0,
        min_size_z=3.0,
        max_size_x=3.0,
        max_size_y=4.0,
        max_size_z=3.0,
        min_source_x=0.5,
        min_source_y=0.5,
        min_source_z=1.8,
        max_source_x=0.5,
        max_source_y=0.5,
        max_source_z=1.8,
        min_mic_distance=0.1,
        max_mic_distance=0.1,
        p=1.0,
    )
    augment(samples=samples, sample_rate=sample_rate)


@pytest.mark.parametrize("num_channels", [1, 2, 3])
def test_multichannel_input(num_channels):
    random.seed(1)
    sample_rate = 16000
    samples = get_sinc_impulse(sample_rate, 10)
    n_channels = np.tile(samples, (num_channels, 1))
    augment = RoomSimulator(leave_length_unchanged=True)
    # Setting the seed is important for reproduction
    np.random.seed(1)
    augmented_samples = augment(samples=samples, sample_rate=sample_rate)

    assert augmented_samples.shape == samples.shape

    augment.freeze_parameters()
    np.random.seed(1)
    augmented_n_channels = augment(samples=n_channels, sample_rate=sample_rate)

    assert augmented_n_channels.shape == n_channels.shape

    assert np.allclose(augmented_samples, augmented_n_channels[0])


@pytest.mark.parametrize("leave_length_unchanged", [True, False])
def test_input_with_absorption(leave_length_unchanged):
    random.seed(1)
    sample_rate = 16000
    samples = get_sinc_impulse(sample_rate, 10)

    augment = RoomSimulator(
        p=1.0,
        leave_length_unchanged=leave_length_unchanged,
    )

    processed_samples = augment(samples=samples, sample_rate=sample_rate)

    # Store a measured rt60 parameter
    theoretical_rt60 = augment.room.rt60_theory()
    measured_rt60 = augment.room.measure_rt60()[0][0]

    # Experimentally set that in this case
    assert np.isclose(theoretical_rt60, measured_rt60, atol=0.065)
    assert processed_samples.dtype == samples.dtype
    assert not np.allclose(processed_samples[: len(samples)], samples)
    assert len(processed_samples.shape) == 1


@pytest.mark.parametrize("leave_length_unchanged", [True, False])
def test_input_with_rt60(leave_length_unchanged):
    random.seed(1)
    sample_rate = 16000
    samples = get_sinc_impulse(sample_rate, 10)

    augment = RoomSimulator(
        p=1.0,
        calculation_mode="rt60",
        min_target_rt60=0.3,
        max_target_rt60=0.3,
        leave_length_unchanged=leave_length_unchanged,
    )

    processed_samples = augment(samples=samples, sample_rate=sample_rate)

    # Store a measured rt60 parameter
    theoretical_rt60 = augment.room.rt60_theory()
    measured_rt60 = augment.room.measure_rt60()[0][0]

    # Experimentally set that in this case. Target t60
    # is expected to deviate quite a bit.
    assert np.isclose(0.3, theoretical_rt60, atol=0.05)
    assert np.isclose(theoretical_rt60, measured_rt60, atol=0.065)
    assert processed_samples.dtype == samples.dtype
    assert not np.allclose(processed_samples[: len(samples)], samples)
    assert len(processed_samples.shape) == 1



import numpy as np
import pytest

from audiomentations import AddGaussianSNR


def test_gaussian_noise_snr_defaults():
    np.random.seed(42)
    samples_in = np.random.normal(0, 1, size=1024).astype(np.float32)
    augmenter = AddGaussianSNR(p=1.0)
    std_in = np.mean(np.abs(samples_in))
    samples_out = augmenter(samples=samples_in, sample_rate=16000)
    std_out = np.mean(np.abs(samples_out))
    assert samples_out.dtype == np.float32
    assert not (float(std_out) == pytest.approx(0.0))
    assert std_out > std_in


def test_gaussian_noise_snr():
    np.random.seed(42)
    samples_in = np.random.normal(0, 1, size=1024).astype(np.float32)
    augmenter = AddGaussianSNR(min_snr_db=15, max_snr_db=35, p=1.0)
    std_in = np.mean(np.abs(samples_in))
    samples_out = augmenter(samples=samples_in, sample_rate=16000)
    std_out = np.mean(np.abs(samples_out))
    assert samples_out.dtype == np.float32
    assert not (float(std_out) == pytest.approx(0.0))
    assert std_out > std_in


def test_gaussian_noise_snr_multichannel():
    np.random.seed(42)
    samples = np.random.normal(0, 0.1, size=(3, 8888)).astype(np.float32)
    augmenter = AddGaussianSNR(min_snr_db=15, max_snr_db=35, p=1.0)
    samples_out = augmenter(samples=samples, sample_rate=16000)

    assert samples_out.dtype == np.float32
    assert float(np.sum(np.abs(samples_out))) > float(np.sum(np.abs(samples)))


def test_validation():
    with pytest.raises(ValueError):
        AddGaussianSNR(min_snr_db=40.0, max_snr_db=20.0)



import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from audiomentations.augmentations.seven_band_parametric_eq import SevenBandParametricEQ


@pytest.mark.parametrize("shape", [(44100,), (1, 22049), (2, 10000)])
def test_apply_eq(shape: tuple):
    samples_in = np.random.normal(0.0, 0.5, size=shape).astype(np.float32)
    sample_rate = 44100
    augmenter = SevenBandParametricEQ(p=1.0)

    samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert samples_out.shape == shape

    with np.testing.assert_raises(AssertionError):
        assert_array_almost_equal(samples_out, samples_in)



import random

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import (
    convert_frequency_to_mel,
    convert_mel_to_frequency,
)


class Aliasing(BaseWaveformTransform):
    """
    Apply an aliasing effect to the audio by downsampling to a lower
    sample rate without filtering and upsampling after that.
    """

    supports_multichannel = True

    def __init__(
        self, min_sample_rate: int = 8000, max_sample_rate: int = 30000, p: float = 0.5
    ):
        """
        :param min_sample_rate: Minimum target sample rate to downsample to
        :param max_sample_rate: Maximum target sample rate to downsample to
        :param p: The probability of applying this transform
        """
        super().__init__(p)

        if min_sample_rate < 2:
            raise ValueError("min_sample_rate must be greater than or equal to 2")

        if min_sample_rate > max_sample_rate:
            raise ValueError("min_sample_rate must not be larger than max_sample_rate")
        
        self.min_sample_rate = min_sample_rate
        self.max_sample_rate = max_sample_rate
        self.min_mel = convert_frequency_to_mel(min_sample_rate)
        self.max_mel = convert_frequency_to_mel(max_sample_rate)

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["new_sample_rate"] = convert_mel_to_frequency(
                random.uniform(self.min_mel, self.max_mel)
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        n = samples.shape[-1]
        x = np.linspace(0, n, num=n)
        dwn_n = round(n * float(self.parameters["new_sample_rate"]) / sample_rate)
        dwn_x = np.linspace(0, n, num=dwn_n)
        if len(samples.shape) > 1:
            distorted_samples = np.zeros((samples.shape[0], n), dtype=np.float32)
            for i in range(samples.shape[0]):
                dwn_samples = np.interp(dwn_x, x, samples[i])
                distorted_samples[i] = np.interp(x, dwn_x, dwn_samples)
        else:
            dwn_samples = np.interp(dwn_x, x, samples)
            distorted_samples = np.interp(x, dwn_x, dwn_samples).astype(np.float32)
        return distorted_samples



import json
import os
import pickle
import random

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal

from audiomentations import AddShortNoises, PolarityInversion
from audiomentations.core.transforms_interface import (
    MultichannelAudioNotSupportedException,
)
from audiomentations.core.utils import calculate_rms
from demo.demo import DEMO_DIR


def test_add_short_noises():
    sample_rate = 44100
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
        np.float32
    )
    rms_before = calculate_rms(samples)
    augmenter = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_time_between_sounds=2.0,
        max_time_between_sounds=8.0,
        p=1.0,
    )
    samples_out = augmenter(samples=samples, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples.shape

    rms_after = calculate_rms(samples_out)
    assert rms_after > rms_before


def test_add_short_noises_with_signal_gain_during_noise():
    sample_rate = 44100
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
        np.float32
    )
    rms_before = calculate_rms(samples)
    augmenter = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_snr_db=50.0,
        max_snr_db=50.0,
        min_time_between_sounds=2.0,
        max_time_between_sounds=4.0,
        signal_gain_db_during_noise=-100,
        p=1.0,
    )
    samples_out = augmenter(samples=samples, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples.shape

    rms_after = calculate_rms(samples_out)
    assert rms_after < rms_before


def test_add_short_noises_with_noise_transform():
    sample_rate = 44100
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
        np.float32
    )
    augmenter = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_snr_db=50.0,
        max_snr_db=50.0,
        min_time_between_sounds=1.0,
        max_time_between_sounds=2.0,
        noise_transform=PolarityInversion(p=1.0),
        p=1.0,
    )
    samples_out_with_noise_transform = augmenter(
        samples=samples, sample_rate=sample_rate
    )
    assert samples_out_with_noise_transform.dtype == np.float32
    assert samples_out_with_noise_transform.shape == samples.shape

    augmenter.noise_transform = None
    samples_out_without_noise_transform = augmenter(
        samples=samples, sample_rate=sample_rate
    )

    with np.testing.assert_raises(AssertionError):
        assert_array_almost_equal(
            samples_out_without_noise_transform, samples_out_with_noise_transform
        )


def test_input_shorter_than_noise():
    """
    Verify correct behavior when the input sound is shorter than the added noise sounds.
    """
    sample_rate = 44100
    samples = np.sin(
        np.linspace(0, 440 * 2 * np.pi, int(0.03125 * sample_rate))
    ).astype(np.float32)
    rms_before = calculate_rms(samples)
    augmenter = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_time_between_sounds=0.00001,
        max_time_between_sounds=0.00002,
        p=1.0,
    )
    samples_out = augmenter(samples=samples, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples.shape

    rms_after = calculate_rms(samples_out)
    assert rms_after > rms_before


def test_add_silence():
    """Check that AddShortNoises does not crash if a noise is completely silent."""
    sample_rate = 48000
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 7000)).astype(np.float32)
    augmenter = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "silence"),
        min_time_between_sounds=0.001,
        max_time_between_sounds=0.002,
        p=1.0,
    )

    augmenter.parameters = {
        "should_apply": True,
        "sounds": [
            {
                "fade_in_time": 0.04257633246838298,
                "start": -0.00013191289693534575,
                "end": 0.8071696744046519,
                "fade_out_time": 0.07119110196424423,
                "file_path": os.path.join(DEMO_DIR, "silence", "silence.wav"),
                "snr_db": 19.040001423519563,
            }
        ],
    }
    augmenter.freeze_parameters()

    samples_out = augmenter(samples=samples, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples.shape


def test_too_long_fade_time():
    """Check that a too long fade time does not result in an exception."""
    sample_rate = 44100
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
        np.float32
    )
    rms_before = calculate_rms(samples)
    augmenter = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_time_between_sounds=2.0,
        max_time_between_sounds=8.0,
        min_fade_in_time=0.9,
        max_fade_in_time=0.99,
        min_fade_out_time=0.9,
        max_fade_out_time=0.99,
        p=1.0,
    )
    samples_out = augmenter(samples=samples, sample_rate=sample_rate)
    assert samples_out.dtype == np.float32
    assert samples_out.shape == samples.shape

    rms_after = calculate_rms(samples_out)
    assert rms_after > rms_before


def test_frozen_parameters():
    sample_rate = 44100
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
        np.float32
    )
    augmenter = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_time_between_sounds=2.0,
        max_time_between_sounds=8.0,
        p=1.0,
    )
    samples_out1 = augmenter(samples=samples, sample_rate=sample_rate)

    augmenter.freeze_parameters()
    samples_out2 = augmenter(samples, sample_rate)

    assert_array_equal(samples_out1, samples_out2)


def test_multichannel_audio_not_supported_yet():
    sample_rate = 44100
    samples_chn0 = np.sin(np.linspace(0, 440 * 2 * np.pi, 2 * sample_rate)).astype(
        np.float32
    )
    samples_chn1 = np.sin(np.linspace(0, 440 * 2 * np.pi, 2 * sample_rate)).astype(
        np.float32
    )
    samples = np.vstack((samples_chn0, samples_chn1))

    augmenter = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_time_between_sounds=2.0,
        max_time_between_sounds=8.0,
        p=1.0,
    )
    with pytest.raises(MultichannelAudioNotSupportedException):
        augmenter(samples=samples, sample_rate=sample_rate)


def test_noise_rms_parameter():
    np.random.seed(80085)
    random.seed("ling")
    sample_rate = 44100
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
        np.float32
    )
    rms_before = calculate_rms(samples)
    augmenter_relative_to_whole_input = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_time_between_sounds=2.0,
        max_time_between_sounds=8.0,
        min_snr_db=5,
        max_snr_db=5,
        noise_rms="relative_to_whole_input",
        p=1.0,
    )

    augmenter_absolute = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_time_between_sounds=2.0,
        max_time_between_sounds=8.0,
        noise_rms="absolute",
        p=1.0,
    )

    samples_out_absolute = augmenter_absolute(samples=samples, sample_rate=sample_rate)
    samples_out_relative_to_whole_input = augmenter_relative_to_whole_input(
        samples=samples, sample_rate=sample_rate
    )
    assert samples_out_absolute.dtype == np.float32
    assert samples_out_absolute.shape == samples.shape
    rms_after_relative_to_whole_path = calculate_rms(
        samples_out_relative_to_whole_input
    )
    rms_after_absolute = calculate_rms(samples_out_absolute)
    assert rms_after_absolute > rms_before
    assert rms_after_relative_to_whole_path > rms_before


def test_include_silence_in_noise_rms_calculation():
    np.random.seed(420)
    random.seed(420)
    sample_rate = 44100
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
        np.float32
    )
    rms_before = calculate_rms(samples)
    augmenter = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_time_between_sounds=2.0,
        max_time_between_sounds=4.0,
        noise_rms="absolute",
        include_silence_in_noise_rms_estimation=False,
        p=1.0,
    )

    samples_out = augmenter(samples=samples, sample_rate=sample_rate)

    rms_after = calculate_rms(samples_out)
    assert rms_after > rms_before


def test_add_noises_with_same_level():
    dummy_audio = np.random.randint(1, 5, 250000)
    transform_same_noise_level = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_snr_db=15,
        max_snr_db=30,
        noise_rms="relative",
        add_all_noises_with_same_level=True,
        min_time_between_sounds=0.5,
        max_time_between_sounds=1,
        p=1,
    )

    transform_different_noise_level = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "short_noises"),
        min_snr_db=15,
        max_snr_db=30,
        add_all_noises_with_same_level=False,
        min_time_between_sounds=0.5,
        max_time_between_sounds=1,
        p=1,
    )

    for i in range(3):
        transform_same_noise_level.randomize_parameters(dummy_audio, sample_rate=44100)
        sounds = transform_same_noise_level.parameters["sounds"]
        snr_sounds_same_level = [sounds[j]["snr_db"] for j in range(len(sounds))]
        transform_different_noise_level.randomize_parameters(
            dummy_audio, sample_rate=44100
        )
        sounds = transform_different_noise_level.parameters["sounds"]
        snr_sounds_different_level = [sounds[j]["snr_db"] for j in range(len(sounds))]

        assert len(set(snr_sounds_same_level)) == 1
        assert len(set(snr_sounds_different_level)) > 1


def test_short_noise_containing_digital_silence_and_include_silence_in_noise_rms_estimation_false():
    # Test a regression where it tried to calculate the RMS of an empty array
    sample_rate = 44100
    samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 1 * sample_rate)).astype(
        np.float32
    )
    noise_path = os.path.join(DEMO_DIR, "tap.wav")
    transform = AddShortNoises(
        sounds_path=os.path.join(DEMO_DIR, "tap.wav"),
        include_silence_in_noise_rms_estimation=False,
        p=1.0,
    )
    transform.freeze_parameters()
    transform.parameters["should_apply"] = True
    transform.parameters["sounds"] = [
        {
            "fade_in_time": 0.05,
            "start": -0.555,
            "end": 0.035,
            "fade_out_time": 0.05,
            "file_path": noise_path,
            "snr_db": -5,
            "rms_db": -20.0,
        }
    ]

    samples_out = transform(samples, sample_rate)
    assert samples_out.shape == samples.shape


def test_validation():
    with pytest.raises(ValueError):
        AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_snr_db=40.0,
            max_snr_db=20.0,
        )



import random
from typing import Literal

import librosa
import numpy as np
from numpy.typing import NDArray
import python_stretch

from audiomentations.core.transforms_interface import BaseWaveformTransform


class TimeStretch(BaseWaveformTransform):
    """Time stretch the signal without changing the pitch"""

    supports_multichannel = True

    def __init__(
        self,
        min_rate: float = 0.8,
        max_rate: float = 1.25,
        leave_length_unchanged: bool = True,
        method: Literal["librosa_phase_vocoder", "signalsmith_stretch"] = "signalsmith_stretch",
        p: float = 0.5,
    ):
        """
        :param min_rate: Minimum time-stretch rate. Values less than 1.0 slow down the audio (reduce the playback speed).
        :param max_rate: Maximum time-stretch rate. Values greater than 1.0 speed up the audio (increase the playback speed).
        :param leave_length_unchanged: If `True`, the output audio will have the same duration as the input audio.
            If `False`, the duration of the output audio will be altered by the time-stretch rate.
        :param method: "librosa_phase_vocoder" or "signalsmith_stretch"
        :param p: The probability of applying this transform.
        """
        super().__init__(p)
        if min_rate < 0.1:
            raise ValueError("min_rate must be >= 0.1")
        if max_rate > 10:
            raise ValueError("max_rate must be <= 10")
        if min_rate > max_rate:
            raise ValueError("min_rate must not be greater than max_rate")

        self.min_rate = min_rate
        self.max_rate = max_rate
        self.leave_length_unchanged = leave_length_unchanged

        if method not in ("librosa_phase_vocoder", "signalsmith_stretch"):
            raise ValueError(
                'method must be set to either "librosa_phase_vocoder" or "signalsmith_stretch"'
            )
        self.method = method

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            """
            If rate > 1, then the signal is sped up.
            If rate < 1, then the signal is slowed down.
            """
            self.parameters["rate"] = random.uniform(self.min_rate, self.max_rate)

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        original_shape = samples.shape
        if self.method == "signalsmith_stretch":
            original_ndim = samples.ndim
            if original_ndim == 1:
                samples = samples[np.newaxis, :]

            stretch = python_stretch.Signalsmith.Stretch()
            stretch.preset(samples.shape[0], sample_rate)
            stretch.setTimeFactor(self.parameters["rate"])
            time_stretched_samples = stretch.process(samples)

            if time_stretched_samples.ndim > original_ndim:
                time_stretched_samples = time_stretched_samples[0]

        else:  # method == "librosa_phase_vocoder"
            try:
                time_stretched_samples = librosa.effects.time_stretch(
                    samples, rate=self.parameters["rate"]
                )
            except librosa.util.exceptions.ParameterError:
                # In librosa<0.9.0 time_stretch doesn't natively support multichannel audio.
                # Here we use a workaround that simply loops over the channels instead.
                # TODO: Remove this workaround when we remove support for librosa<0.9.0
                time_stretched_channels = []
                for i in range(samples.shape[0]):
                    time_stretched_samples_ch = librosa.effects.time_stretch(
                        samples[i], rate=self.parameters["rate"]
                    )
                    time_stretched_channels.append(time_stretched_samples_ch)
                time_stretched_samples = np.array(
                    time_stretched_channels, dtype=samples.dtype
                )

        if self.leave_length_unchanged:
            # Apply zero padding if the time stretched audio is not long enough to fill the
            # whole space, or crop the time stretched audio if it ended up too long.
            padded_samples = np.zeros(shape=original_shape, dtype=samples.dtype)
            window = time_stretched_samples[..., : samples.shape[-1]]
            actual_window_length = window.shape[-1]
            padded_samples[..., :actual_window_length] = window
            time_stretched_samples = padded_samples

        return time_stretched_samples



from typing import Literal

import python_stretch
import random
import warnings

import librosa
import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class PitchShift(BaseWaveformTransform):
    """Pitch shift the sound up or down without changing the tempo"""

    supports_multichannel = True

    def __init__(
        self,
        min_semitones: float = -4.0,
        max_semitones: float = 4.0,
        method: Literal[
            "librosa_phase_vocoder", "signalsmith_stretch"
        ] = "signalsmith_stretch",
        p: float = 0.5,
    ):
        """
        :param min_semitones: Minimum semitones to shift. A negative number means shift down.
        :param max_semitones: Maximum semitones to shift. A positive number means shift up.
        :param method:
            "librosa_phase_vocoder": slow, low quality, supports any number of channels
            "signalsmith_stretch" (default): fast, high quality, only supports mono and stereo
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        if min_semitones < -24:
            raise ValueError("min_semitones must be >= -24")
        if max_semitones > 24:
            raise ValueError("max_semitones must be <= 24")
        if min_semitones > max_semitones:
            raise ValueError("min_semitones must not be greater than max_semitones")
        self.min_semitones = min_semitones
        self.max_semitones = max_semitones
        if method not in ("librosa_phase_vocoder", "signalsmith_stretch"):
            raise ValueError(
                'method must be set to either "librosa_phase_vocoder" or'
                ' "signalsmith_stretch"'
            )
        self.method = method

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["num_semitones"] = random.uniform(
                self.min_semitones, self.max_semitones
            )

    def apply(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        if self.method == "signalsmith_stretch":
            original_ndim = samples.ndim
            if original_ndim == 1:
                samples = samples[np.newaxis, :]

            stretch = python_stretch.Signalsmith.Stretch()
            stretch.preset(samples.shape[0], sample_rate)
            stretch.setTransposeSemitones(self.parameters["num_semitones"])
            samples = stretch.process(samples)
            if samples.ndim > original_ndim:
                samples = samples[0]
            return samples

        try:
            resample_type = (
                "kaiser_best" if librosa.__version__.startswith("0.8.") else "soxr_hq"
            )
            pitch_shifted_samples = librosa.effects.pitch_shift(
                samples,
                sr=sample_rate,
                n_steps=self.parameters["num_semitones"],
                res_type=resample_type,
            )
        except librosa.util.exceptions.ParameterError:
            warnings.warn(
                "Warning: You are probably using an old version of librosa. Upgrade"
                " librosa to 0.9.0 or later for better performance when applying"
                " PitchShift to stereo audio."
            )
            # In librosa<0.9.0 pitch_shift doesn't natively support multichannel audio.
            # Here we use a workaround that simply loops over the channels instead.
            # TODO: Remove this workaround when we remove support for librosa<0.9.0
            pitch_shifted_samples = np.copy(samples)
            for i in range(samples.shape[0]):
                pitch_shifted_samples[i] = librosa.effects.pitch_shift(
                    pitch_shifted_samples[i],
                    sr=sample_rate,
                    n_steps=self.parameters["num_semitones"],
                )

        return pitch_shifted_samples



import numpy as np
from numpy.testing import assert_almost_equal

from audiomentations import PolarityInversion, Compose


def test_polarity_inversion():
    samples = np.array([1.0, 0.5, -0.25, -0.125, 0.0], dtype=np.float32)
    sample_rate = 16000

    augment = Compose([PolarityInversion(p=1.0)])
    inverted_samples = augment(samples=samples, sample_rate=sample_rate)
    assert_almost_equal(
        inverted_samples, np.array([-1.0, -0.5, 0.25, 0.125, 0.0], dtype=np.float32)
    )
    assert inverted_samples.dtype == np.float32


def test_polarity_inversion_multichannel():
    samples = np.array(
        [[1.0, 0.5, -0.25, -0.125, 0.0], [1.0, 0.5, -0.25, -0.125, 0.0]],
        dtype=np.float32,
    )
    sample_rate = 16000

    augment = Compose([PolarityInversion(p=1.0)])
    inverted_samples = augment(samples=samples, sample_rate=sample_rate)
    assert_almost_equal(
        inverted_samples,
        np.array(
            [[-1.0, -0.5, 0.25, 0.125, 0.0], [-1.0, -0.5, 0.25, 0.125, 0.0]],
            dtype=np.float32,
        ),
    )
    assert inverted_samples.dtype == np.float32



import math

import numpy as np
import pytest

from audiomentations import Resample


@pytest.mark.parametrize(
    "samples",
    [
        np.zeros((512,), dtype=np.float32),
        np.zeros(
            (
                2,
                2512,
            ),
            dtype=np.float32,
        ),
    ],
)
def test_resample(samples):
    sample_rate = 16000
    augmenter = Resample(min_sample_rate=8000, max_sample_rate=44100, p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert samples.dtype == np.float32
    assert samples.shape[-1] <= math.ceil(samples.shape[-1] * 44100 / sample_rate)
    assert samples.shape[-1] >= math.ceil(samples.shape[-1] * 8000 / sample_rate)



import numpy as np
from numpy.testing import assert_array_equal

from audiomentations import Normalize


def test_normalize_positive_peak():
    samples = np.array([0.5, 0.6, -0.2, 0.0], dtype=np.float32)
    sample_rate = 16000
    augmenter = Normalize(p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert np.amax(samples) == 1.0
    assert samples.dtype == np.float32
    assert samples.shape[-1] == 4


def test_normalize_negative_peak():
    samples = np.array([0.5, 0.6, -0.8, 0.0], dtype=np.float32)
    sample_rate = 16000
    augmenter = Normalize(p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert np.amin(samples) == -1.0
    assert samples[-1] == 0.0
    assert samples.dtype == np.float32
    assert samples.shape[-1] == 4


def test_normalize_all_zeros():
    samples = np.array([0.0, 0.0, 0.0, 0.0], dtype=np.float32)
    sample_rate = 16000
    augmenter = Normalize(p=1.0)
    samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert np.amin(samples) == 0.0
    assert samples[-1] == 0.0
    assert samples.dtype == np.float32
    assert samples.shape[-1] == 4


def test_normalize_multichannel():
    samples = np.array(
        [[0.9, 0.5, -0.25, -0.125, 0.0], [0.95, 0.5, -0.25, -0.125, 0.0]],
        dtype=np.float32,
    )
    sample_rate = 16000
    augmenter = Normalize(p=1.0)
    processed_samples = augmenter(samples=samples, sample_rate=sample_rate)

    assert_array_equal(processed_samples, samples / 0.95)
    assert processed_samples.dtype == np.float32


def test_normalize_multichannel_conditionally():
    sample_rate = 16000
    augmenter = Normalize(apply_to="only_too_loud_sounds", p=1.0)

    samples = np.array(
        [[0.9, 0.5, -0.25, -0.125, 0.0], [0.95, 0.5, -0.25, -0.125, 0.0]],
        dtype=np.float32,
    )
    processed_samples = augmenter(samples=samples, sample_rate=sample_rate)
    assert_array_equal(processed_samples, samples)
    assert processed_samples.dtype == np.float32

    samples_too_loud = np.array(
        [[0.9, 0.5, -0.25, -0.125, 0.0], [1.2, 0.5, -0.25, -0.125, 0.0]],
        dtype=np.float32,
    )
    processed_samples = augmenter(samples=samples_too_loud, sample_rate=sample_rate)
    assert_array_equal(processed_samples, samples_too_loud / 1.2)
    assert processed_samples.dtype == np.float32



import random

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import calculate_desired_noise_rms, calculate_rms


class AddGaussianSNR(BaseWaveformTransform):
    """
    Add gaussian noise to the input. A random Signal to Noise Ratio (SNR) will be picked
    uniformly in the Decibel scale. This aligns with human hearing, which is more
    logarithmic than linear.
    """

    supports_multichannel = True

    def __init__(
        self,
        min_snr_db: float = 5.0,
        max_snr_db: float = 40.0,
        p: float = 0.5,
    ):
        """
        :param min_snr_db: Minimum signal-to-noise ratio in dB. A lower number means more noise.
        :param max_snr_db: Maximum signal-to-noise ratio in dB. A greater number means less noise.
        :param p: The probability of applying this transform
        """
        super().__init__(p)

        if min_snr_db > max_snr_db:
            raise ValueError("min_snr_db must not be greater than max_snr_db")
        self.min_snr_db = min_snr_db
        self.max_snr_db = max_snr_db

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            # Pick SNR in decibel scale
            snr = random.uniform(self.min_snr_db, self.max_snr_db)

            clean_rms = calculate_rms(samples)
            noise_rms = calculate_desired_noise_rms(clean_rms=clean_rms, snr=snr)

            # In gaussian noise, the RMS gets roughly equal to the std
            self.parameters["noise_std"] = float(noise_rms)

    def apply(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        noise = np.random.normal(
            0.0, self.parameters["noise_std"], size=samples.shape
        ).astype(np.float32)
        return samples + noise



from __future__ import annotations

import inspect
import random
import warnings
from typing import Any

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.utils import is_waveform_multichannel
from audiomentations.core.serialization import (
    Serializable,
    SerializableMeta,
    get_shortest_class_fullname,
)
from audiomentations.core.utils import format_args


class MultichannelAudioNotSupportedException(Exception):
    pass


class MonoAudioNotSupportedException(Exception):
    pass


class WrongMultichannelAudioShape(Exception):
    pass


class CombinedMeta(SerializableMeta):
    pass


class BaseTransform(Serializable, metaclass=CombinedMeta):
    supports_mono = True
    supports_multichannel = False

    def __init__(self, p=0.5):
        assert 0 <= p <= 1
        self.p = p
        self.parameters = {"should_apply": None}
        self.are_parameters_frozen = False

    def __repr__(self) -> str:
        state = self.get_base_init_args()
        state.update(self.get_transform_init_args())
        return f"{self.__class__.__name__}({format_args(state)})"

    def serialize_parameters(self):
        """Return the parameters as a JSON-serializable dict."""
        return self.parameters

    def freeze_parameters(self):
        """
        Mark all parameters as frozen, i.e. do not randomize them for each call. This can be
        useful if you want to apply an effect with the exact same parameters to multiple sounds.
        """
        self.are_parameters_frozen = True

    def unfreeze_parameters(self):
        """
        Unmark all parameters as frozen, i.e. let them be randomized for each call.
        """
        self.are_parameters_frozen = False

    @classmethod
    def get_class_fullname(cls) -> str:
        return get_shortest_class_fullname(cls)

    @classmethod
    def is_serializable(cls) -> bool:
        return True

    def get_base_init_args(self) -> dict[str, Any]:
        """Returns base init args - p"""
        return {"p": self.p}

    def get_transform_init_args(self) -> dict[str, Any]:
        """Exclude seed from init args during serialization"""
        init_signature = inspect.signature(self.__init__)
        args = {}
        for k, _ in init_signature.parameters.items():
            attr = getattr(self, k, None)
            if attr is not None:
                args[k] = attr
            else:
                warnings.warn(
                    f"Warning: attribute {k} is not found in the transform definition and it won't be printed."
                )
        args.pop("seed", None)  # Remove seed from args
        return args


class BaseWaveformTransform(BaseTransform):
    def apply(self, samples: NDArray[np.float32], sample_rate: int):
        raise NotImplementedError

    def is_multichannel(self, samples):
        return is_waveform_multichannel(samples)

    def __call__(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        if samples.dtype == np.float64:
            warnings.warn(
                "Warning: input samples dtype is np.float64. Converting to np.float32"
            )
            samples = np.float32(samples)
        if not self.are_parameters_frozen or self.parameters["should_apply"] is None:
            self.randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"] and len(samples) > 0:
            if self.is_multichannel(samples):
                # Note: We multiply by 8 here to allow big batches of very short audio
                if samples.shape[0] > samples.shape[1] * 8:
                    raise WrongMultichannelAudioShape(
                        "Multichannel audio must have channels first, not channels"
                        " last. In other words, the shape must be (channels, samples),"
                        " not (samples, channels). See"
                        " https://iver56.github.io/audiomentations/guides/multichannel_audio_array_shapes/"
                        " for more info."
                    )
                if not self.supports_multichannel:
                    raise MultichannelAudioNotSupportedException(
                        "{} only supports mono audio, not multichannel audio. In other"
                        " words, a 1-dimensional input ndarray was expected, but the"
                        " input had more than 1 dimension.".format(
                            self.__class__.__name__
                        )
                    )
            elif not self.supports_mono:
                raise MonoAudioNotSupportedException(
                    "{} only supports multichannel audio, not mono audio".format(
                        self.__class__.__name__
                    )
                )
            return self.apply(samples, sample_rate)
        return samples

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        self.parameters["should_apply"] = random.random() < self.p

    @classmethod
    def get_class_fullname(cls) -> str:
        return get_shortest_class_fullname(cls)

    def to_dict_private(self) -> dict[str, Any]:
        state = {"__class_fullname__": self.get_class_fullname()}
        state.update(self.get_base_init_args())
        state.update(self.get_transform_init_args())
        return state



import random

import numpy as np
from numpy.typing import NDArray
from scipy.signal import butter, sosfilt, sosfiltfilt, sosfilt_zi

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import (
    convert_frequency_to_mel,
    convert_mel_to_frequency,
)


class BaseButterworthFilter(BaseWaveformTransform):
    """
    A `scipy.signal.butter`-based generic filter class.
    """

    supports_multichannel = True

    # The types below must be equal to the ones accepted by
    # the `btype` argument of `scipy.signal.butter`
    ALLOWED_ONE_SIDE_FILTER_TYPES = ("lowpass", "highpass")
    ALLOWED_TWO_SIDE_FILTER_TYPES = ("bandpass", "bandstop")
    ALLOWED_FILTER_TYPES = ALLOWED_ONE_SIDE_FILTER_TYPES + ALLOWED_TWO_SIDE_FILTER_TYPES

    def __init__(self, **kwargs):
        assert "p" in kwargs
        assert "min_rolloff" in kwargs
        assert "max_rolloff" in kwargs
        assert "filter_type" in kwargs
        assert "zero_phase" in kwargs

        super().__init__(kwargs["p"])

        self.filter_type = kwargs["filter_type"]
        self.min_rolloff = kwargs["min_rolloff"]
        self.max_rolloff = kwargs["max_rolloff"]
        self.zero_phase = kwargs["zero_phase"]

        if self.zero_phase:
            assert (
                self.min_rolloff % 12 == 0
            ), "Zero phase filters can only have a steepness which is a multiple of 12 dB/octave"
            assert (
                self.max_rolloff % 12 == 0
            ), "Zero phase filters can only have a steepness which is a multiple of 12 dB/octave"
        else:
            assert (
                self.min_rolloff % 6 == 0
            ), "Non zero phase filters can only have a steepness which is a multiple of 6 dB/octave"
            assert (
                self.max_rolloff % 6 == 0
            ), "Non zero phase filters can only have a steepness which is a multiple of 6 dB/octave"

        assert (
            self.filter_type in BaseButterworthFilter.ALLOWED_FILTER_TYPES
        ), "Filter type must be one of: " + ", ".join(
            BaseButterworthFilter.ALLOWED_FILTER_TYPES
        )

        assert ("min_cutoff_freq" in kwargs and "max_cutoff_freq" in kwargs) or (
            "min_center_freq" in kwargs
            and "max_center_freq" in kwargs
            and "min_bandwidth_fraction" in kwargs
            and "max_bandwidth_fraction" in kwargs
        ), "Arguments for either a one-sided, or a two-sided filter must be given"

        if "min_cutoff_freq" in kwargs:
            self.initialize_one_sided_filter(
                min_cutoff_freq=kwargs["min_cutoff_freq"],
                max_cutoff_freq=kwargs["max_cutoff_freq"],
            )
        elif "min_center_freq" in kwargs:
            self.initialize_two_sided_filter(
                min_center_freq=kwargs["min_center_freq"],
                max_center_freq=kwargs["max_center_freq"],
                min_bandwidth_fraction=kwargs["min_bandwidth_fraction"],
                max_bandwidth_fraction=kwargs["max_bandwidth_fraction"],
            )

    def initialize_one_sided_filter(
        self,
        min_cutoff_freq,
        max_cutoff_freq,
    ):
        """
        :param min_cutoff_freq: Minimum cutoff frequency in hertz
        :param max_cutoff_freq: Maximum cutoff frequency in hertz
        :param min_rolloff: Minimum filter roll-off (in dB/octave).
            Must be a multiple of 6
        :param max_rolloff: Maximum filter roll-off (in dB/octave)
            Must be a multiple of 6
        :param p: The probability of applying this transform
        """

        self.min_cutoff_freq = min_cutoff_freq
        self.max_cutoff_freq = max_cutoff_freq
        if self.min_cutoff_freq > self.max_cutoff_freq:
            raise ValueError("min_cutoff_freq must not be greater than max_cutoff_freq")

        if self.min_rolloff < 6 or self.min_rolloff % 6 != 0:
            raise ValueError(
                "min_rolloff must be 6 or greater, as well as a multiple of 6 (e.g. 6, 12, 18, 24...)"
            )
        if self.max_rolloff < 6 or self.max_rolloff % 6 != 0:
            raise ValueError(
                "max_rolloff must be 6 or greater, as well as a multiple of 6 (e.g. 6, 12, 18, 24...)"
            )
        if self.min_rolloff > self.max_rolloff:
            raise ValueError("min_rolloff must not be greater than max_rolloff")

    def initialize_two_sided_filter(
        self,
        min_center_freq,
        max_center_freq,
        min_bandwidth_fraction,
        max_bandwidth_fraction,
    ):
        """
        :param min_center_freq: Minimum center frequency in hertz
        :param max_center_freq: Maximum center frequency in hertz
        :param min_bandwidth_fraction: Minimum bandwidth fraction relative to center
            frequency (number between 0.0 and 2.0)
        :param max_bandwidth_fraction: Maximum bandwidth fraction relative to center
            frequency (number between 0.0 and 2.0)
        :param min_rolloff: Minimum filter roll-off (in dB/octave).
            Must be a multiple of 6
        :param max_rolloff: Maximum filter roll-off (in dB/octave)
            Must be a multiple of 6
        :param p: The probability of applying this transform
        """

        self.min_center_freq = min_center_freq
        self.max_center_freq = max_center_freq
        self.min_bandwidth_fraction = min_bandwidth_fraction
        self.max_bandwidth_fraction = max_bandwidth_fraction

        if self.min_center_freq > self.max_center_freq:
            raise ValueError("min_center_freq must not be greater than max_center_freq")
        if self.min_bandwidth_fraction <= 0.0:
            raise ValueError("min_bandwidth_fraction must be a positive number")
        if self.max_bandwidth_fraction >= 2.0:
            raise ValueError(
                "max_bandwidth_fraction should be smaller than 2.0, since otherwise"
                " the low cut frequency of the band can be smaller than 0 Hz."
            )
        if self.min_bandwidth_fraction > self.max_bandwidth_fraction:
            raise ValueError(
                "min_bandwidth_fraction must not be greater than max_bandwidth_fraction"
            )

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int = None):
        super().randomize_parameters(samples, sample_rate)
        if self.zero_phase:
            random_order = random.randint(
                self.min_rolloff // 12, self.max_rolloff // 12
            )
            self.parameters["rolloff"] = random_order * 12
        else:
            random_order = random.randint(self.min_rolloff // 6, self.max_rolloff // 6)
            self.parameters["rolloff"] = random_order * 6

        if self.filter_type in BaseButterworthFilter.ALLOWED_ONE_SIDE_FILTER_TYPES:
            cutoff_mel = np.random.uniform(
                low=convert_frequency_to_mel(self.min_cutoff_freq),
                high=convert_frequency_to_mel(self.max_cutoff_freq),
            )
            self.parameters["cutoff_freq"] = convert_mel_to_frequency(cutoff_mel)
        elif self.filter_type in BaseButterworthFilter.ALLOWED_TWO_SIDE_FILTER_TYPES:
            center_mel = np.random.uniform(
                low=convert_frequency_to_mel(self.min_center_freq),
                high=convert_frequency_to_mel(self.max_center_freq),
            )
            self.parameters["center_freq"] = convert_mel_to_frequency(center_mel)

            bandwidth_fraction = np.random.uniform(
                low=self.min_bandwidth_fraction, high=self.max_bandwidth_fraction
            )
            self.parameters["bandwidth"] = (
                self.parameters["center_freq"] * bandwidth_fraction
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int = None) -> NDArray[np.float32]:
        assert samples.dtype == np.float32

        if self.filter_type in BaseButterworthFilter.ALLOWED_ONE_SIDE_FILTER_TYPES:
            cutoff_freq = self.parameters["cutoff_freq"]
            nyquist_freq = sample_rate // 2
            if cutoff_freq > nyquist_freq:
                # Ensure that the cutoff frequency does not exceed the Nyquist
                # frequency to avoid an exception from scipy
                cutoff_freq = nyquist_freq * 0.9999
            sos = butter(
                self.parameters["rolloff"] // (12 if self.zero_phase else 6),
                cutoff_freq,
                btype=self.filter_type,
                analog=False,
                fs=sample_rate,
                output="sos",
            )
        elif self.filter_type in BaseButterworthFilter.ALLOWED_TWO_SIDE_FILTER_TYPES:
            low_freq = self.parameters["center_freq"] - self.parameters["bandwidth"] / 2
            high_freq = (
                self.parameters["center_freq"] + self.parameters["bandwidth"] / 2
            )
            nyquist_freq = sample_rate // 2
            if high_freq > nyquist_freq:
                # Ensure that the upper critical frequency does not exceed the Nyquist
                # frequency to avoid an exception from scipy
                high_freq = nyquist_freq * 0.9999
            sos = butter(
                self.parameters["rolloff"] // (12 if self.zero_phase else 6),
                [low_freq, high_freq],
                btype=self.filter_type,
                analog=False,
                fs=sample_rate,
                output="sos",
            )

        # The actual processing takes place here
        if len(samples.shape) == 1:
            if self.zero_phase:
                processed_samples = sosfiltfilt(sos, samples)
            else:
                processed_samples, _ = sosfilt(
                    sos, samples, zi=sosfilt_zi(sos) * samples[0]
                )
            processed_samples = processed_samples.astype(np.float32)
        else:
            processed_samples = np.zeros_like(samples, dtype=np.float32)
            if self.zero_phase:
                for chn_idx in range(samples.shape[0]):
                    processed_samples[chn_idx, :] = sosfiltfilt(
                        sos, samples[chn_idx, :]
                    )
            else:
                zi = sosfilt_zi(sos)
                for chn_idx in range(samples.shape[0]):
                    processed_samples[chn_idx, :], _ = sosfilt(
                        sos, samples[chn_idx, :], zi=zi * samples[chn_idx, 0]
                    )

        return processed_samples



from copy import deepcopy

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal

from audiomentations import AdjustDuration, Gain, RepeatPart, Shift
from audiomentations.core.utils import get_crossfade_mask_pair
from tests.test_peaking_filter import get_chirp_test


def adapt_ndim(samples, ndim):
    if samples.ndim < ndim:
        samples = samples[np.newaxis, :]
    return samples


def assert_high_frequency_energy_absence(
    audio_array,
    sample_rate,
    frequency_threshold: float = 10_000.0,
    energy_threshold: float = 0.15,
):
    spectrum = np.fft.rfft(audio_array)
    freqs = np.fft.rfftfreq(len(audio_array), 1 / sample_rate)
    assert np.all(
        np.abs(spectrum[freqs > frequency_threshold]) < energy_threshold
    ), "Detected energy in frequencies above the given frequency_threshold!"


def test_replace_one_repeat():
    augment = RepeatPart(mode="replace", crossfade_duration=0.0, p=1.0)
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": 3,
        "repeats": 1,
        "part_start_index": 1,
    }
    augment.freeze_parameters()

    for ndim in (1, 2):
        samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            dtype=np.float32,
        )
        samples = adapt_ndim(samples, ndim)
        processed_samples = augment(samples=samples, sample_rate=4000)
        target_samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.7, 0.8, 0.9, 1.0],
            dtype=np.float32,
        )
        target_samples = adapt_ndim(target_samples, ndim)
        assert_array_almost_equal(processed_samples, target_samples)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32


def test_replace_one_repeat_transformed():
    augment = RepeatPart(
        mode="replace",
        crossfade_duration=0.0,
        part_transform=Gain(min_gain_db=-6.0, max_gain_db=-6.0, p=1.0),
        p=1.0,
    )
    part_gain_factor = 0.5011872336272722
    augment.part_transform.parameters = {
        "should_apply": True,
        "amplitude_ratio": part_gain_factor,  # -6 dB
    }
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": 3,
        "repeats": 1,
        "part_start_index": 1,
    }
    augment.freeze_parameters()

    for ndim in (1, 2):
        samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            dtype=np.float32,
        )
        samples = adapt_ndim(samples, ndim)

        processed_samples = augment(samples=samples, sample_rate=4000)

        target_samples = np.array(
            [
                0.0,
                0.1,
                0.2,
                0.3,
                0.1 * part_gain_factor,
                0.2 * part_gain_factor,
                0.3 * part_gain_factor,
                0.7,
                0.8,
                0.9,
                1.0,
            ],
            dtype=np.float32,
        )
        target_samples = adapt_ndim(target_samples, ndim)
        assert_array_almost_equal(processed_samples, target_samples)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32


def test_freeze_and_unfreeze_part_transform_parameters():
    augment = RepeatPart(
        min_part_duration=0.1,
        max_part_duration=0.2,
        part_transform=Gain(p=1.0),
        crossfade_duration=0.0,
        p=1.0,
    )
    dummy_samples = np.zeros(40, dtype=np.float32)
    augment(dummy_samples, 40)
    params1 = deepcopy(augment.part_transform.parameters)
    augment.freeze_parameters()
    augment(dummy_samples, 40)
    params2 = deepcopy(augment.part_transform.parameters)
    assert params1 == params2

    augment.unfreeze_parameters()
    augment(dummy_samples, 40)
    params3 = deepcopy(augment.part_transform.parameters)
    assert params3 != params2


def test_replace_one_repeat_near_end():
    augment = RepeatPart(mode="replace", crossfade_duration=0.0, p=1.0)
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": 3,
        "repeats": 1,
        "part_start_index": 7,
    }
    augment.freeze_parameters()

    for ndim in (1, 2):
        samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            dtype=np.float32,
        )
        samples = adapt_ndim(samples, ndim)
        processed_samples = augment(samples=samples, sample_rate=4000)
        target_samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.7],
            dtype=np.float32,
        )
        target_samples = adapt_ndim(target_samples, ndim)
        assert_array_almost_equal(processed_samples, target_samples)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32


def test_replace_two_repeats():
    augment = RepeatPart(mode="replace", crossfade_duration=0.0, p=1.0)
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": 3,
        "repeats": 2,
        "part_start_index": 1,
    }
    augment.freeze_parameters()

    for ndim in (1, 2):
        samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            dtype=np.float32,
        )
        samples = adapt_ndim(samples, ndim)
        processed_samples = augment(samples=samples, sample_rate=4000)
        target_samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 1.0],
            dtype=np.float32,
        )
        target_samples = adapt_ndim(target_samples, ndim)
        assert_array_almost_equal(processed_samples, target_samples)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32


def test_replace_many_repeats_exceed_input_length():
    augment = RepeatPart(mode="replace", crossfade_duration=0.0, p=1.0)
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": 3,
        "repeats": 9,
        "part_start_index": 1,
    }
    augment.freeze_parameters()

    for ndim in (1, 2):
        samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            dtype=np.float32,
        )
        samples = adapt_ndim(samples, ndim)
        processed_samples = augment(samples=samples, sample_rate=4000)
        target_samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1],
            dtype=np.float32,
        )
        target_samples = adapt_ndim(target_samples, ndim)
        assert_array_almost_equal(processed_samples, target_samples)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32


def test_replace_many_transformed_repeats_exceed_input_length():
    augment = RepeatPart(
        mode="replace",
        crossfade_duration=0.0,
        part_transform=Gain(min_gain_db=-6.0, max_gain_db=-6.0, p=1.0),
        p=1.0,
    )
    part_gain_factor = 0.5011872336272722
    augment.part_transform.parameters = {
        "should_apply": True,
        "amplitude_ratio": part_gain_factor,  # -6 dB
    }
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": 3,
        "repeats": 9,
        "part_start_index": 1,
    }
    augment.freeze_parameters()

    for ndim in (1, 2):
        samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            dtype=np.float32,
        )
        samples = adapt_ndim(samples, ndim)
        processed_samples = augment(samples=samples, sample_rate=4000)
        target_samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1],
            dtype=np.float32,
        )
        target_samples[4:] *= part_gain_factor
        target_samples = adapt_ndim(target_samples, ndim)
        assert_array_almost_equal(processed_samples, target_samples)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32


def test_insert_one_repeat():
    augment = RepeatPart(mode="insert", crossfade_duration=0.0, p=1.0)
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": 3,
        "repeats": 1,
        "part_start_index": 1,
    }
    augment.freeze_parameters()

    samples = np.array(
        [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0], dtype=np.float32
    )
    processed_samples = augment(samples=samples, sample_rate=4000)
    assert_array_almost_equal(
        processed_samples,
        np.array(
            [0.0, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            dtype=np.float32,
        ),
    )
    assert processed_samples.dtype == np.float32


def test_insert_two_repeats():
    augment = RepeatPart(mode="insert", crossfade_duration=0.0, p=1.0)
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": 3,
        "repeats": 2,
        "part_start_index": 0,
    }
    augment.freeze_parameters()

    samples = np.array([0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7], dtype=np.float32)
    processed_samples = augment(samples=samples, sample_rate=4000)
    assert_array_almost_equal(
        processed_samples,
        np.array(
            [0.0, 0.1, 0.2, 0.0, 0.1, 0.2, 0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7],
            dtype=np.float32,
        ),
    )
    assert processed_samples.dtype == np.float32


def test_insert_two_elongated_repeats():
    augment = RepeatPart(
        mode="insert",
        crossfade_duration=0.0,
        part_transform=AdjustDuration(duration_samples=6, p=1.0),
        p=1.0,
    )
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": 3,
        "repeats": 2,
        "part_start_index": 0,
    }
    augment.freeze_parameters()

    samples = np.array([0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7], dtype=np.float32)
    processed_samples = augment(samples=samples, sample_rate=4000)
    assert_array_almost_equal(
        processed_samples,
        np.array(
            [
                0.0,
                0.1,
                0.2,
                0.0,  # first repetition starts here
                0.1,
                0.2,
                0.0,  # start of zero padding
                0.0,
                0.0,  # end of zero padding
                0.0,  # second repetition starts here
                0.1,
                0.2,
                0.0,  # start of zero padding
                0.0,
                0.0,  # end of zero padding
                0.3,
                0.4,
                0.5,
                0.6,
                0.7,
            ],
            dtype=np.float32,
        ),
    )
    assert processed_samples.dtype == np.float32


def test_insert_two_repeats_with_crossfading():
    augment = RepeatPart(mode="insert", crossfade_duration=0.005, p=1.0)
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": int(0.25 * 44100),
        "repeats": 2,
        "part_start_index": 150,
    }
    augment.freeze_parameters()

    sample_rate = 44100
    samples = get_chirp_test(sample_rate, duration=2) * 0.3
    processed_samples = augment(samples=samples, sample_rate=sample_rate)

    assert processed_samples.dtype == np.float32


def test_insert_two_repeats_with_crossfading_and_part_transform():
    # The purpose of this test is to cover two different cases:
    # 1) The fade from the last part to the tail is correlated (use equal-gain fade)
    # 2) The fade from the last part to the tail is uncorrelated (use equal-energy fade)
    for part_transform in [Gain(p=0.0), Shift(p=1.0)]:
        augment = RepeatPart(
            mode="insert",
            crossfade_duration=0.005,
            part_transform=part_transform,
            p=1.0,
        )
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": int(0.25 * 44100),
            "repeats": 2,
            "part_start_index": 150,
        }
        augment.freeze_parameters()

        sample_rate = 44100
        samples = get_chirp_test(sample_rate, duration=2) * 0.3
        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        assert processed_samples.dtype == np.float32


def test_replace_mode_two_repeats_with_crossfading():
    augment = RepeatPart(mode="replace", crossfade_duration=0.005, p=1.0)
    part_num_samples = int(0.25 * 44100)
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": part_num_samples,
        "repeats": 2,
        "part_start_index": 142,
    }
    augment.freeze_parameters()

    sample_rate = 44100
    amplitude = 0.3
    samples = get_chirp_test(sample_rate, duration=2) * amplitude
    processed_samples = augment(samples=samples, sample_rate=sample_rate)

    assert processed_samples.dtype == np.float32

    augment_without_crossfade = RepeatPart(
        mode="replace", crossfade_duration=0.000, p=1.0
    )
    augment_without_crossfade.parameters = augment.parameters
    augment_without_crossfade.freeze_parameters()
    processed_samples_without_crossfade = augment_without_crossfade(
        samples=samples, sample_rate=sample_rate
    )

    idx0 = augment.parameters["part_start_index"] + part_num_samples - 1
    idx1 = idx0 + 1
    seam_impulse_magnitude_without_crossfade = abs(
        processed_samples_without_crossfade[..., idx1]
        - processed_samples_without_crossfade[..., idx0]
    )
    assert seam_impulse_magnitude_without_crossfade > amplitude  # audible click!
    seam_impulse_magnitude_with_crossfade = abs(
        processed_samples[..., idx1] - processed_samples[..., idx0]
    )
    assert seam_impulse_magnitude_with_crossfade < 0.11 * amplitude  # smooth

    assert processed_samples.shape == processed_samples_without_crossfade.shape


def test_replace_mode_repeats_with_crossfading_at_end():
    # Test that it doesn't crash if the end of the array is in the middle of
    # the crossfade of the last part (in replace mode)
    augment = RepeatPart(mode="replace", crossfade_duration=0.124, p=1.0)
    part_num_samples = int(0.29 * 44100)
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": part_num_samples,
        "repeats": 5,
        "part_start_index": int(0.22 * 44100),
    }
    augment.freeze_parameters()

    sample_rate = 44100
    amplitude = 0.3
    samples = get_chirp_test(sample_rate, duration=2) * amplitude
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert processed_samples.shape == samples.shape


def test_replace_mode_many_repeats_with_crossfading():
    # Test that it doesn't fade in some of the original audio at the end when there
    # enough repetitions to fill the whole array when in replace mode.
    augment = RepeatPart(mode="replace", crossfade_duration=0.124, p=1.0)
    part_num_samples = int(0.27 * 44100)
    augment.parameters = {
        "should_apply": True,
        "part_num_samples": part_num_samples,
        "repeats": 20,
        "part_start_index": int(0.22 * 44100),
    }
    augment.freeze_parameters()

    sample_rate = 44100
    amplitude = 0.3
    samples = get_chirp_test(sample_rate, duration=2) * amplitude
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert processed_samples.shape == samples.shape

    # The very last part should not have any crossfade into the original signal,
    # where there is only high-freq audio.
    assert_high_frequency_energy_absence(
        processed_samples[int(1.9 * 44100) :], sample_rate
    )


def test_invalid_parameters():
    with pytest.raises(ValueError):
        RepeatPart(crossfade_duration=-0.0001)
    with pytest.raises(ValueError):
        RepeatPart(crossfade_duration=0.0001)
    with pytest.raises(ValueError):
        RepeatPart(min_part_duration=0.0001)
    with pytest.raises(ValueError):
        RepeatPart(min_repeats=0, max_repeats=1)
    with pytest.raises(ValueError):
        RepeatPart(min_repeats=2, max_repeats=1)
    with pytest.raises(ValueError):
        RepeatPart(min_part_duration=0.5, max_part_duration=0.2)
    with pytest.raises(ValueError):
        RepeatPart(min_part_duration=0.01, crossfade_duration=0.1)
    with pytest.raises(ValueError):
        RepeatPart(mode="append")


def test_warn_too_short_input():
    augmenter = RepeatPart(
        mode="replace",
        min_part_duration=0.1,
        max_part_duration=0.1,
        crossfade_duration=0.0,
        p=1.0,
    )
    samples = np.random.uniform(low=-0.5, high=0.5, size=(100,)).astype(np.float32)
    with pytest.warns(
        UserWarning,
        match=(
            "The input sound is not long enough for applying the RepeatPart"
            " transform with the current parameters."
        ),
    ):
        processed_samples = augmenter(samples, sample_rate=10_000)

    assert_array_equal(processed_samples, samples)


def test_very_short_crossfade_mask_pair():
    fade_in, fade_out = get_crossfade_mask_pair(2)
    assert_array_almost_equal(fade_in, np.array([0.0, 1.0], dtype=np.float32))
    assert_array_almost_equal(fade_out, np.array([1.0, 0.0], dtype=np.float32))



import numpy as np
from numpy.typing import NDArray

from audiomentations import LowShelfFilter, PeakingFilter, HighShelfFilter
from audiomentations.core.transforms_interface import BaseWaveformTransform


class SevenBandParametricEQ(BaseWaveformTransform):
    """
    Adjust the volume of different frequency bands. This transform is a 7-band
    parametric equalizer - a combination of one low shelf filter, five peaking filters
    and one high shelf filter, all with randomized gains, Q values and center frequencies.

    Because this transform changes the timbre but keeps the overall "class" of the
    sound the same (depending on the application), it can be used for data augmentation to
    make ML models more robust to various frequency spectra. Many things can affect
    the spectrum, for example:

    * the nature and quality of the sound source
    * room acoustics
    * any objects between the microphone and the sound source
    * microphone type or model
    * the distance between the sound source and the microphone

    The seven bands have center frequencies picked in the following ranges (min-max):

    * 42-95 Hz
    * 91-204 Hz
    * 196-441 Hz
    * 421-948 Hz
    * 909-2045 Hz
    * 1957-4404 Hz
    * 4216-9486 Hz
    """

    supports_multichannel = True

    def __init__(
        self,
        min_gain_db: float = -12.0,
        max_gain_db: float = 12.0,
        p: float = 0.5,
    ):
        """
        :param min_gain_db: Minimum number of dB to cut or boost a band
        :param max_gain_db: Maximum number of dB to cut or boost a band
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        assert min_gain_db <= max_gain_db
        
        self.min_gain_db = min_gain_db
        self.max_gain_db = max_gain_db

        self.low_shelf_filter = LowShelfFilter(
            min_center_freq=42.0,
            max_center_freq=95.0,
            min_gain_db=min_gain_db,
            max_gain_db=max_gain_db,
            p=1.0,
        )
        self.peaking_filters = [
            PeakingFilter(
                min_center_freq=91.0,
                max_center_freq=204.0,
                min_gain_db=min_gain_db,
                max_gain_db=max_gain_db,
                p=1.0,
            ),
            PeakingFilter(
                min_center_freq=196.0,
                max_center_freq=441.0,
                min_gain_db=min_gain_db,
                max_gain_db=max_gain_db,
                p=1.0,
            ),
            PeakingFilter(
                min_center_freq=421.0,
                max_center_freq=948.0,
                min_gain_db=min_gain_db,
                max_gain_db=max_gain_db,
                p=1.0,
            ),
            PeakingFilter(
                min_center_freq=909.0,
                max_center_freq=2045.0,
                min_gain_db=min_gain_db,
                max_gain_db=max_gain_db,
                p=1.0,
            ),
            PeakingFilter(
                min_center_freq=1957.0,
                max_center_freq=4404.0,
                min_gain_db=min_gain_db,
                max_gain_db=max_gain_db,
                p=1.0,
            ),
        ]
        self.high_shelf_filter = HighShelfFilter(
            min_center_freq=4216.0,
            max_center_freq=9486.0,
            min_gain_db=min_gain_db,
            max_gain_db=max_gain_db,
            p=1.0,
        )
        self.low_shelf_filter.freeze_parameters()
        for i in range(len(self.peaking_filters)):
            self.peaking_filters[i].freeze_parameters()
        self.high_shelf_filter.freeze_parameters()

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        self.low_shelf_filter.randomize_parameters(samples, sample_rate)
        for i in range(len(self.peaking_filters)):
            self.peaking_filters[i].randomize_parameters(samples, sample_rate)
        self.high_shelf_filter.randomize_parameters(samples, sample_rate)

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        samples = self.low_shelf_filter(samples, sample_rate)
        for i in range(len(self.peaking_filters)):
            samples = self.peaking_filters[i](samples, sample_rate)
        samples = self.high_shelf_filter(samples, sample_rate)
        return samples



import random

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class AddGaussianNoise(BaseWaveformTransform):
    """Add gaussian noise to the samples"""

    supports_multichannel = True

    def __init__(self, min_amplitude=0.001, max_amplitude=0.015, p=0.5):
        """

        :param min_amplitude: Minimum noise amplification factor
        :param max_amplitude: Maximum noise amplification factor
        :param p:
        """
        super().__init__(p)
        assert min_amplitude > 0.0
        assert max_amplitude > 0.0
        assert max_amplitude >= min_amplitude
        self.min_amplitude = min_amplitude
        self.max_amplitude = max_amplitude

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["amplitude"] = random.uniform(
                self.min_amplitude, self.max_amplitude
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        noise = np.random.randn(*samples.shape).astype(np.float32)
        samples = samples + self.parameters["amplitude"] * noise
        return samples
